# pylint: disable=too-many-lines
from typing import List, Tuple, Callable
import io
import json
from uuid import uuid4
from pathlib import Path
import defusedxml.ElementTree as ET

import numpy as np
import xtgeo
import dash
import dash_table
from dash.dependencies import Input, Output, State
import dash_html_components as html
import dash_core_components as dcc
import dash_bootstrap_components as dbc
import webviz_core_components as wcc
import webviz_subsurface_components

from webviz_config.webviz_assets import WEBVIZ_ASSETS
from webviz_config import WebvizPluginABC
from webviz_config.webviz_store import webvizstore

import webviz_subsurface
from webviz_subsurface._datainput.surface import get_surface_layers
from webviz_subsurface._datainput.well import get_well_layers
from ._huv_xsection import HuvXsection
from ._huv_table import FilterTable


class HorizonUncertaintyViewer(WebvizPluginABC):
    """Visualizes depth uncertainty for surfaces in map view and cross section view.

    The cross section is defined by wellfiles and surfacefiles or a polyline.
    Polylines are drawn interactivly in map view.

    !> The plugin reads information from a COHIBA model file.

    * **`basedir`:** Path to folder with model_file.xml.
       Make sure that the folder has the same format as a COHIBA folder.
    * **`planned_wells_dir`:** Path to folder with planned well files.
       Make sure that all planned wells have format 'ROXAR RMS well'."""

    # pylint: disable=too-many-locals,too-many-instance-attributes
    def __init__(
        self,
        app: dash.Dash,
        basedir: Path,
        planned_wells_dir: Path = None,
    ):

        super().__init__()
        self.plotly_theme = app.webviz_settings["theme"].plotly_theme
        self.uid = uuid4()
        WEBVIZ_ASSETS.add(
            Path(webviz_subsurface.__file__).parent / "_assets" / "css" / "modal.css"
        )
        self.set_callbacks(app)

        self.basedir = basedir
        self.planned_wells_dir = planned_wells_dir
        self.modelfile_path = basedir / "model_file.xml"
        self.modelfile = get_path(self.modelfile_path)
        self.surfaces = load_surfaces(basedir, self.modelfile_path)
        self.planned_wellfiles = (
            json.load(find_files(planned_wells_dir, "*.txt"))
            if planned_wells_dir
            else None
        )
        self.wellfiles = json.load(find_files(basedir / "input" / "welldata", "*.txt"))
        self.wellfiles = [str(get_path(Path(w))) for w in self.wellfiles]
        self.allfiles = json.load(find_files(basedir))
        self.allfiles.append(self.modelfile_path)
        self.allfiles += self.planned_wellfiles
        self.planned_wellfiles = [
            str(get_path(Path(w))) for w in self.planned_wellfiles
        ]
        self.surface_attributes = {}
        for i, surface in enumerate(self.surfaces):
            self.surface_attributes[surface["name"]] = {
                "color": get_color(i),
                "order": i,
                "name": surface["name"],
                "topofzone": surface["topofzone"],
                "surface": surface["d_"],
                "surface_de": surface["de_"],
                "surface_dt": surface["dt_"],
                "surface_dr": surface["dr_"],
                "surface_dte": surface["dte_"],
            }

        self.surfacenames = [surface["name"] for surface in self.surfaces]
        # Log files
        zonation_status_file = get_zonation_status(basedir)
        well_points_file = get_well_points(basedir)
        zonelog_name = get_zonelog_name(self.modelfile)
        self.xsec = HuvXsection(
            self.surface_attributes,
            zonation_status_file,
            well_points_file,
            zonelog_name,
        )
        target_points_file = get_target_points(basedir)
        self.df_well_target_points = FilterTable(target_points_file, well_points_file)

        # Wellfiles and planned wells
        self.planned_wells = {}
        if planned_wells_dir is not None:
            self.planned_wells = {wf: xtgeo.Well(wf) for wf in self.planned_wellfiles}

        self.wells = {wf: xtgeo.Well(wf) for wf in self.wellfiles}

        # Store current layers
        self.state = {"switch": False}
        self.layers_state = []

    def ids(self, element: str) -> str:
        return f"{element}-id-{self.uid}"

    @property
    def cross_section_graph_layout(self) -> html.Div:
        return html.Div(
            children=[
                wcc.Graph(
                    id=self.ids("xsec-view"),
                )
            ]
        )

    @property
    def cross_section_widgets_layout(self) -> html.Div:
        return html.Div(
            children=[
                html.Div(
                    children=[
                        dbc.Button(
                            "Surface Settings",
                            id=self.ids("button-open-graph-settings"),
                            color="light",
                            className="mr-1",
                        ),
                        dbc.Modal(
                            children=[
                                dbc.ModalHeader("Surface Settings"),
                                dbc.ModalBody(
                                    children=[
                                        html.Label(
                                            style={
                                                "font-weight": "bold",
                                                "textAlign": "Left",
                                            },
                                            children="Select Surfaces",
                                        ),
                                        dcc.Checklist(
                                            id=self.ids("all-surfaces-checkbox"),
                                            options=[{"label": "all", "value": "True"}],
                                            value=["True"],
                                            persistence=True,
                                            persistence_type="session",
                                        ),
                                        dcc.Checklist(
                                            id=self.ids("surfaces-checklist"),
                                            options=[
                                                {"label": name, "value": name}
                                                for name in self.surfacenames
                                            ],
                                            value=self.surfacenames,
                                            persistence=True,
                                            persistence_type="session",
                                        ),
                                        dcc.Checklist(
                                            id=self.ids("surfaces-de-checklist"),
                                            options=[
                                                {
                                                    "label": name + " SD",
                                                    "value": name,
                                                    "disabled": False,
                                                }
                                                for name in self.surfacenames
                                            ],
                                            value=self.surfacenames,
                                            persistence=True,
                                            persistence_type="session",
                                        ),
                                    ],
                                ),
                                dbc.ModalFooter(
                                    children=[
                                        dbc.Button(
                                            "Close",
                                            id=self.ids("button-close-graph-settings"),
                                            className="ml-auto",
                                        ),
                                        dbc.Button(
                                            "Apply changes",
                                            id=self.ids("button-apply-checklist"),
                                            className="ml-auto",
                                        ),
                                    ]
                                ),
                            ],
                            id=self.ids("modal-graph-settings"),
                            size="sm",
                            centered=True,
                            backdrop=False,
                            fade=False,
                        ),
                        dbc.Button(
                            "Well Settings", id=self.ids("button-open-well-settings")
                        ),
                        dbc.Modal(
                            children=[
                                dbc.ModalHeader("Well Settings"),
                                dbc.ModalBody(
                                    children=[
                                        html.Label(
                                            style={
                                                "font-weight": "bold",
                                                "textAlign": "Left",
                                            },
                                            children="Select Well Attributes",
                                        ),
                                        dcc.Checklist(
                                            id=self.ids("all-well-settings-checkbox"),
                                            options=[{"label": "all", "value": "True"}],
                                            value=["True"],
                                            persistence=True,
                                            persistence_type="session",
                                        ),
                                        dcc.Checklist(
                                            id=self.ids("well-settings-checklist"),
                                            options=[
                                                {
                                                    "label": "Zonelog",
                                                    "value": "zonelog",
                                                },
                                                {
                                                    "label": "Zonation points",
                                                    "value": "zonation_points",
                                                },
                                                {
                                                    "label": "Conditional points",
                                                    "value": "conditional_points",
                                                },
                                            ],
                                            value=[
                                                "zonelog",
                                                "zonation_points",
                                                "conditional_points",
                                            ],
                                            persistence=True,
                                            persistence_type="session",
                                        ),
                                    ],
                                ),
                                dbc.ModalFooter(
                                    children=[
                                        dbc.Button(
                                            "Close",
                                            id=self.ids("button-close-well-settings"),
                                            className="ml-auto",
                                        ),
                                        dbc.Button(
                                            "Apply",
                                            id=self.ids(
                                                "button-apply-well-settings-checklist"
                                            ),
                                            className="ml-auto",
                                        ),
                                    ]
                                ),
                            ],
                            id=self.ids("modal-well-settings"),
                            size="sm",
                            centered=True,
                            backdrop=False,
                            fade=False,
                        ),
                    ],
                ),
                wcc.FlexBox(
                    children=[
                        html.Div(
                            children=[
                                html.Label(
                                    style={
                                        "font-weight": "bold",
                                        "textAlign": "center",
                                    },
                                    children="Select well",
                                ),
                                dcc.Dropdown(
                                    id=self.ids("well-dropdown"),
                                    options=[
                                        {
                                            "label": self.wells[wf].wellname,
                                            "value": str(wf),
                                        }
                                        for wf in self.wellfiles
                                    ]
                                    + [
                                        {
                                            "label": self.planned_wells[wf].wellname,
                                            "value": str(wf),
                                        }
                                        for wf in self.planned_wellfiles
                                    ],
                                    value=str(self.wellfiles[0]),
                                    clearable=False,
                                    disabled=False,
                                    persistence=True,
                                    persistence_type="session",
                                ),
                            ]
                        ),
                    ],
                ),
                html.Div(
                    children=[
                        html.Div(
                            style={
                                "marginTop": "0px",
                                "height": "800px",
                                "zIndex": -9999,
                            },
                            children=[self.cross_section_graph_layout],
                            id=self.ids("cross-section-view"),
                        )
                    ]
                ),
            ]
        )

    @property
    def target_points_tab_layout(self) -> dash_table.DataTable:
        df = self.df_well_target_points.get_targetpoints_df()
        return dash_table.DataTable(
            id=self.ids("target-point-table"),
            columns=[{"name": i, "id": i} for i in df.columns],
            data=df.to_dict("records"),
            sort_action="native",
            filter_action="native",
        )

    @property
    def well_points_tab_layout(self) -> html.Div:
        return html.Div(
            [
                dbc.Button("Table Settings", id=self.ids("button-open-table-settings")),
                dbc.Modal(
                    children=[
                        dbc.ModalHeader("Table Settings"),
                        dbc.ModalBody(
                            children=[
                                html.Label(
                                    style={
                                        "font-weight": "bold",
                                        "textAlign": "Left",
                                    },
                                    children="Select Table Columns",
                                ),
                                dcc.Checklist(
                                    id=self.ids("columns-checklist"),
                                    options=[
                                        {"label": name, "value": column_name}
                                        for name, column_name in zip(
                                            self.df_well_target_points.get_wellpoints_df()
                                            .keys()
                                            .values,
                                            self.df_well_target_points.get_wellpoints_df()
                                            .keys()
                                            .values,
                                        )
                                    ],
                                    value=[
                                        "Surface",
                                        "Well",
                                        "TVD",
                                        "MD",
                                        "Outlier",
                                        "Deleted",
                                        "Residual",
                                    ],
                                    persistence=True,
                                    persistence_type="session",
                                ),
                            ],
                        ),
                        dbc.ModalFooter(
                            children=[
                                dbc.Button(
                                    "Close",
                                    id=self.ids("button-close-table-settings"),
                                    className="ml-auto",
                                ),
                                dbc.Button(
                                    "Apply",
                                    id=self.ids("button-apply-columnlist"),
                                    className="ml-auto",
                                ),
                            ]
                        ),
                    ],
                    id=self.ids("modal-table-settings"),
                    size="sm",
                    centered=True,
                    backdrop=False,
                    fade=False,
                ),
                html.Div(id=self.ids("well-points-table-container")),
            ]
        )

    @property
    def left_flexbox_layout(self) -> html.Div:
        return html.Div(
            children=[
                wcc.FlexBox(
                    children=[
                        dcc.RadioItems(
                            labelStyle={"display": "inline-block"},
                            options=[
                                {"label": "Map view", "value": "map-view"},
                                {"label": "Surface picks table", "value": "table-view"},
                            ],
                            id=self.ids("map-table-radioitems"),
                            value="map-view",
                            persistence=True,
                            persistence_type="session",
                        )
                    ],
                    style={"padding": "5px"},
                ),
                html.Div(
                    id=self.ids("hidden-div-map-view"), children=[self.map_view_layout]
                ),  # Hidden div to store polyline points when in table view
                html.Div(
                    id=self.ids("hidden-div-table-view"),
                    children=[self.table_view_layout],
                ),
            ]
        )

    @property
    def map_view_layout(self) -> html.Div:
        return html.Div(
            children=[
                wcc.FlexBox(
                    children=[
                        html.Div(
                            children=[
                                html.Label(
                                    style={
                                        "font-weight": "bold",
                                        "textAlign": "center",
                                    },
                                    children="Select surface",
                                ),
                                dcc.Dropdown(
                                    id=self.ids("map-dropdown"),
                                    options=[
                                        {"label": name, "value": name}
                                        for name in self.surfacenames
                                    ],
                                    value=self.surfacenames[0],
                                    clearable=False,
                                    persistence=True,
                                    persistence_type="session",
                                ),
                            ]
                        ),
                    ],
                ),
                html.Div(
                    style={
                        "marginTop": "0px",
                        "height": "800px",
                        "zIndex": -9999,
                    },
                    children=[
                        # pylint: disable=no-member
                        webviz_subsurface_components.LeafletMap(
                            id=self.ids("layered-map"),
                            layers=[],
                            unitScale={},
                            autoScaleMap=True,
                            minZoom=-5,
                            drawTools={
                                "drawMarker": False,
                                "drawPolygon": False,
                                "drawPolyline": True,
                                "position": "topright",
                            },
                            switch={
                                "value": self.state["switch"],
                                "disabled": False,
                                "label": "Hillshading",
                            },
                            mouseCoords={"position": "bottomright"},
                            colorBar={"position": "bottomright"},
                        ),
                    ],
                ),
            ]
        )

    @property
    def table_view_layout(self) -> html.Div:
        df = self.xsec.get_intersection_dataframe(self.wells[self.wellfiles[0]])
        return html.Div(
            children=[
                html.Label(
                    id=self.ids("surface-picks-label"),
                    style={
                        "font-weight": "bold",
                        "textAlign": "center",
                    },
                ),
                dash_table.DataTable(
                    id=self.ids("uncertainty-table"),
                    columns=[{"name": i, "id": i} for i in df.columns],
                    data=df.to_dict("records"),
                    sort_action="native",
                    filter_action="native",
                ),
            ],
            style={"marginTop": "0px"},
        )

    @property
    def layout(self) -> dcc.Tabs:
        return dcc.Tabs(
            children=[
                dcc.Tab(
                    label="Cross section & map view",
                    children=[
                        wcc.FlexBox(
                            id=self.ids("layout"),
                            children=[
                                html.Div(
                                    style={"flex": 1}, children=self.left_flexbox_layout
                                ),
                                html.Div(
                                    style={"flex": 1.5},
                                    children=self.cross_section_widgets_layout,
                                ),
                            ],
                        )
                    ],
                ),
                dcc.Tab(
                    label="Target Points",
                    children=[html.Div(children=self.target_points_tab_layout)],
                ),
                dcc.Tab(label="Well Points", children=[self.well_points_tab_layout]),
            ]
        )

    # pylint: disable=too-many-statements
    def set_callbacks(self, app: dash.Dash) -> None:
        @app.callback(
            Output(self.ids("layered-map"), "layers"),
            [
                Input(self.ids("map-dropdown"), "value"),
                Input(self.ids("layered-map"), "switch"),  # Toggle hillshading on/off
                Input(self.ids("well-dropdown"), "value"),  # Wellfile
            ],
        )
        def _render_map(surface_name, switch, wellfile):
            """Renders map view for one surface with de, dt, dte, dr, dre and depth
            Wells marked with circles, trajectory and hillshading toggle
            """
            # Store layers when switching to hillshading
            if self.state["switch"] is not switch["value"]:
                hillshade_layers = self.layers_state.copy()
                for layer in hillshade_layers:
                    if "shader" in layer["data"][0]:
                        layer["data"][0]["shader"]["type"] = (
                            "hillshading" if switch["value"] is True else None
                        )
                        layer["action"] = "update"
                self.state["switch"] = switch["value"]
                return hillshade_layers

            surfaces = [
                self.surface_attributes[surface_name]["surface"],
                self.surface_attributes[surface_name]["surface_de"],
                self.surface_attributes[surface_name]["surface_dr"],
                self.surface_attributes[surface_name]["surface_dt"],
                self.surface_attributes[surface_name]["surface_dte"],
            ]
            well_layers = get_well_layers(
                self.wells,
                self.planned_wells,
                surface_name,
                surfaces[0],
                wellfile,
            )
            layers = get_surface_layers(switch, surface_name, surfaces)
            layers.extend(well_layers)
            # Deletes old layers when switching surface in dropdown
            old_layers = self.layers_state
            self.layers_state = layers.copy()
            if old_layers is not None and len(old_layers) > 0:
                for layer in old_layers:
                    layer["action"] = "delete"
                old_layers.extend(layers)
                layers = old_layers
            return layers

        @app.callback(
            Output(self.ids("xsec-view"), "figure"),
            [
                Input(self.ids("button-apply-checklist"), "n_clicks"),
                Input(self.ids("button-apply-well-settings-checklist"), "n_clicks"),
                Input(self.ids("well-dropdown"), "value"),  # wellpath
                Input(self.ids("layered-map"), "polyline_points"),  # Polyline
            ],
            [
                State(self.ids("surfaces-checklist"), "value"),  # surfacefiles
                State(self.ids("surfaces-de-checklist"), "value"),  # surfacefiles keys
                State(self.ids("well-settings-checklist"), "value"),  # Well settings
            ],
        )
        def _render_xsection(
            n_apply_sfc,
            n_apply_well,
            wellfile,
            polyline,
            surfacefiles,
            de_keys,
            well_settings,
        ):
            """ Renders cross section view from wellfile or polyline drawn in map view """
            _ = n_apply_sfc, n_apply_well
            ctx = dash.callback_context
            if wellfile in self.wellfiles:
                well = self.wells[wellfile]
                is_planned = False
            else:
                well = self.planned_wells[wellfile]
                is_planned = True
            well.create_relative_hlen()
            if (
                ctx.triggered[0]["prop_id"]
                == self.ids("layered-map") + ".polyline_points"
                and polyline is not None
            ):
                well = None
            self.xsec.set_de_and_surface_lines(surfacefiles, de_keys, well, polyline)
            self.xsec.set_xsec_fig(
                surfacefiles, de_keys, well_settings, well, is_planned=is_planned
            )
            return self.xsec.fig

        @app.callback(
            Output(self.ids("well-dropdown"), "value"),  # Wellfile
            [Input(self.ids("layered-map"), "clicked_shape")],  # Clicked circle on map
        )
        def _update_render_map_and_xsection(clicked_shape):
            """ Updates well dropdown value when choosing a new circle """
            return (
                clicked_shape["id"] if clicked_shape is not None else self.wellfiles[0]
            )

        @app.callback(
            Output(self.ids("surfaces-checklist"), "value"),
            [Input(self.ids("all-surfaces-checkbox"), "value")],
        )
        def _update_surface_tickboxes(all_surfaces_checkbox):
            """ Toggle on/off all surfaces in graph settings modal """
            return self.surfacenames if all_surfaces_checkbox == ["True"] else []

        @app.callback(
            Output(self.ids("modal-graph-settings"), "is_open"),
            [
                Input(self.ids("button-open-graph-settings"), "n_clicks"),
                Input(self.ids("button-close-graph-settings"), "n_clicks"),
                Input(self.ids("button-open-graph-settings"), "disabled"),
            ],
            [State(self.ids("modal-graph-settings"), "is_open")],
        )
        def _toggle_modal_graph_settings(n_open, n_close, disabled, is_open):
            """ Open or close graph settings modal button """
            if disabled:
                switch = False
            elif n_open or n_close:
                switch = not is_open
            else:
                switch = is_open
            return switch

        @app.callback(
            Output(self.ids("surfaces-de-checklist"), "options"),
            [Input(self.ids("surfaces-checklist"), "value")],
            [State(self.ids("surfaces-de-checklist"), "options")],
        )
        def _disable_error_checkboxes(surface_values, de_options):
            """Removes ability to toggle depth error when
            corresponding surface is disabled in graph settings modal
            """
            for i, opt in enumerate(de_options):
                if (surface_values is None) or (opt["value"] not in surface_values):
                    de_options[i]["disabled"] = True
                else:
                    de_options[i]["disabled"] = False
            return de_options

        @app.callback(
            Output(self.ids("well-settings-checklist"), "value"),
            [Input(self.ids("all-well-settings-checkbox"), "value")],
        )
        def _update_well_settings_tickboxes(all_well_attributes_checkbox):
            """ Toggle on/off all options in well settings modal """
            return (
                ["zonelog", "zonation_points", "conditional_points"]
                if all_well_attributes_checkbox == ["True"]
                else []
            )

        @app.callback(
            Output(self.ids("modal-well-settings"), "is_open"),
            [
                Input(self.ids("button-open-well-settings"), "n_clicks"),
                Input(self.ids("button-close-well-settings"), "n_clicks"),
            ],
            [State(self.ids("modal-well-settings"), "is_open")],
        )
        def _toggle_modal_well_settings(n_open, n_close, is_open):
            """ Open or close well settings modal button """
            if n_open or n_close:
                return not is_open
            return is_open

        @app.callback(
            Output(self.ids("well-points-table-container"), "children"),
            [
                Input(self.ids("button-apply-columnlist"), "n_clicks"),
            ],
            [
                State(self.ids("columns-checklist"), "value"),
            ],  # columns list
        )
        # pylint: disable=unused-variable, unused-argument
        def display_output(n_clicks, column_list):
            """ Renders wellpoints table from csv file """
            wellpoints_df = self.df_well_target_points.update_wellpoints_df(column_list)
            return html.Div(
                [
                    dash_table.DataTable(
                        id=self.ids("well-points-table"),
                        columns=[{"name": i, "id": i} for i in wellpoints_df.columns],
                        data=wellpoints_df.to_dict("records"),
                        sort_action="native",
                        filter_action="native",
                    ),
                ]
            )

        @app.callback(
            Output(self.ids("modal-table-settings"), "is_open"),
            [
                Input(self.ids("button-open-table-settings"), "n_clicks"),
                Input(self.ids("button-close-table-settings"), "n_clicks"),
            ],
            [State(self.ids("modal-table-settings"), "is_open")],
        )
        def _toggle_modal_table_settings(n_open, n_close, is_open):
            """ Open or close table settings modal button """
            if n_open or n_close:
                return not is_open
            return is_open

        @app.callback(
            [
                Output(self.ids("hidden-div-map-view"), "hidden"),
                Output(self.ids("hidden-div-table-view"), "hidden"),
            ],
            [Input(self.ids("map-table-radioitems"), "value")],
        )
        def _toggle_left_flexbox_content(value):
            """ Returns which left flexbox content is visible/hidden """
            if value == "table-view":
                switch = True, False
            else:
                switch = False, True
            return switch

        @app.callback(
            Output(self.ids("uncertainty-table"), "data"),
            [Input(self.ids("well-dropdown"), "value")],
        )
        def _render_uncertainty_table(wellfile):
            if wellfile in self.wellfiles:
                well = self.wells[wellfile]
            else:
                well = self.planned_wells[wellfile]
            df = self.xsec.get_intersection_dataframe(well)
            return df.to_dict("records")

        @app.callback(
            Output(self.ids("surface-picks-label"), "children"),
            [Input(self.ids("well-dropdown"), "value")],
        )
        def _render_surface_picks_label(wellfile):
            if wellfile in self.wellfiles:
                wellname = self.wells[wellfile].wellname
            else:
                wellname = self.planned_wells[wellfile].wellname
            return f"Surface picks for {wellname}"

    def add_webvizstore(self) -> List[Tuple[Callable, list]]:
        functions = [(get_path, [{"path": Path(fn)}]) for fn in self.allfiles]
        functions.append((find_files, [{"folder": self.basedir}]))
        if self.planned_wells_dir is not None:
            functions.append(
                (
                    find_files,
                    [
                        {"folder": self.planned_wells_dir, "pattern": "*.txt"},
                        {
                            "folder": self.basedir / "input" / "welldata",
                            "pattern": "*.txt",
                        },
                    ],
                )
            )
        functions.append(
            (
                get_surfaces,
                [
                    {
                        "basedir": self.basedir,
                        "modelfile": self.modelfile_path,
                    }
                ],
            )
        )
        return functions


@webvizstore
def get_path(path) -> Path:
    return Path(path)


def get_color(i):
    """Create a list of colors for surface layers
    Args:
        i: Index of surface layer in surfacefiles list
    Returns:
        List of colors for surface layers
    """
    colors = [
        "rgb(70,130,180)",  # Steel blue
        "rgb(0,0,255)",  # Blue
        "rgb(173,255,47)",  # Green yellow
        "rgb(0,128,0)",  # Green
        "rgb(0,255,0)",  # Lime
        "rgb(60,179,113)",  # Medium sea green
        "rgb(255,105,180)",  # Pink
        "rgb(221,160,221)",  # Plum
        "rgb(255,255,0)",  # Yellow
        "rgb(244,164,96)",  # Tan
        "rgb(255,140,0)",  # Orange
        "rgb(255,69,0)",  # Blood orange
        "rgb(255,0,0)",  # Red
        "rgb(220,20,60)",  # Crimson
        "rgb(255,0,255)",  # Fuchsia
    ]
    n_colors = len(colors)
    return colors[(i) % (n_colors)]


def load_surfaces(basedir: Path, modelfile):
    surface_types = {
        "depth": "d_",
        "depth-trend": "dt_",
        "depth-error": "de_",
        "depth-trend-error": "dte_",
        "depth-residual": "dr_",
    }
    surface_data = json.load(get_surfaces(basedir, modelfile))
    for surface in surface_data:
        for stype in surface_types.values():
            surface[stype] = surface_from_json(surface[stype]) if stype else None
    return surface_data


@webvizstore
def get_surfaces(basedir: Path, modelfile) -> io.BytesIO:
    modelfile = ET.parse(get_path(modelfile)).getroot()
    surface_types = {
        "depth": "d_",
        "depth-trend": "dt_",
        "depth-error": "de_",
        "depth-trend-error": "dte_",
        "depth-residual": "dr_",
    }
    surface_wrappers = modelfile.findall(".//surface")
    surfaces = []
    for element in surface_wrappers:
        surface = {
            "name": element.findtext("name"),
            "topofzone": element.findtext("top-of-zone"),
        }
        for s_mtype, s_type in surface_types.items():
            surface[s_type] = (
                surface_to_json(
                    str(
                        basedir
                        / "output"
                        / "surfaces"
                        / f"{s_type}{surface['name']}.rxb"
                    )
                )
                if element.find("output").findtext(s_mtype) == "yes"
                else None
            )
        surfaces.append(surface)
    return io.BytesIO(json.dumps(surfaces).encode())


def surface_to_json(surfacepath: Path) -> str:
    surface = xtgeo.surface_from_file(str(surfacepath), fformat="irap_binary")
    return json.dumps(
        {
            "ncol": surface.ncol,
            "nrow": surface.nrow,
            "xori": surface.xori,
            "yori": surface.yori,
            "rotation": surface.rotation,
            "xinc": surface.xinc,
            "yinc": surface.yinc,
            "values": surface.values.copy().filled(np.nan).tolist(),
        }
    )


def surface_from_json(surfaceobj):
    data = json.loads(surfaceobj)
    surface = xtgeo.RegularSurface(**data)
    surface.values = np.array(data["values"])
    return surface


# TODO(Sigurd) Delete unused function?
# def extract_topofzone_names(modelfile):
#    modelfile = ET.parse(modelfile).getroot()
#    surface_wrappers = modelfile.findall(".//surface")
#    topofzone_names = []
#    for element in surface_wrappers:
#        name = element.findtext("top-of-zone")
#        topofzone_names.append(name)
#    return topofzone_names


# TODO(Sigurd) Delete unused function?
# def get_well_files(basedir: Path) -> io.BytesIO:
#    well_dir = os.path.join(basedir, "input", "welldata")
#    well_dir = basedir / "input" / "welldata"
#    well_files = []
#    try:
#        for file in os.listdir(well_dir):
#            if Path(file).suffix == ".txt":
#                well_files.append(os.path.join(well_dir, file))
#    except FileNotFoundError:
#        pass
#    well_files.sort()
#    return io.BytesIO(json.dumps(well_files).encode())


def get_target_points(basedir: Path) -> Path:
    return get_path(basedir / "output" / "log_files" / "targetpoints.csv")


def get_well_points(basedir: Path) -> Path:
    return get_path(basedir / "output" / "log_files" / "wellpoints.csv")


def get_zonelog_name(modelfile):
    modelfile = ET.parse(modelfile).getroot()
    zonelog_wrapper = modelfile.findtext(".//zone-log-name")
    return zonelog_wrapper


def get_zonation_status(basedir: Path) -> Path:
    return get_path(basedir / "output" / "log_files" / "zonation_status.csv")


# TODO(Sigurd) Delete unused?
# @webvizstore
# def get_surface_files(basedir: Path, surface_names, surface_type) -> io.BytesIO:
#    surface_dir = basedir / "output" / "surfaces"
#    surface_files = [
#        surface_dir / f"{surface_type}{surface_name}.rxb"
#        for surface_name in surface_names
#    ]
#    for path in surface_files:
#        if not path.is_file():
#            surface_files = None
#    if surface_files is not None:
#        surface_files = [str(surf) for surf in surface_files]
#    return io.BytesIO(json.dumps(surface_files).encode())


@webvizstore
def find_files(folder: Path, pattern: str = "*") -> io.BytesIO:
    return io.BytesIO(
        json.dumps(
            sorted(
                [
                    str(filename)
                    for filename in folder.glob(f"**/{pattern}")
                    if filename.is_file()
                ]
            )
        ).encode()
    )



from typing import Dict, Union

import pandas as pd


def update_errorplot(
    df: pd.DataFrame, enscolors: Dict[str, str]
) -> Dict[
    str, Union[list, dict]
]:  # TODO(Ruben Thoms) Better to make use of a TypedDict here at some point.

    df["RFT_NAME"] = df.agg(
        lambda x: f"{x['WELL']} {int(x['YEAR'])} {x['ZONE']} ({int(x['TVD'])} TVD)",
        axis=1,
    )
    df["DIFFMEAN"] = df.groupby(["WELL", "DATE", "ZONE", "TVD", "ENSEMBLE"])[
        "ABSDIFF"
    ].transform("median")
    traces = []
    for i, (ensemble, ensdf) in enumerate(df.groupby("ENSEMBLE")):
        if i == 0:
            ensdf = ensdf.sort_values(by=["DIFFMEAN"])
        traces.append(
            {
                "x": ensdf["DIFF"],
                "y": ensdf["RFT_NAME"],
                "name": ensemble,
                "type": "box",
                "orientation": "h",
                "offsetgroup": i,
                "marker": {"color": enscolors[ensemble]},
            }
        )
    layout = {
        "margin": {"l": 250, "r": 0, "b": 50, "t": 100},
        "height": 750,
        "legend": {"orientation": "h"},
        "boxmode": "group",
        "xaxis": {"title": "Difference in bar"},
    }

    return {"data": traces, "layout": layout}



from typing import Union, TYPE_CHECKING

import dash_html_components as html
import dash_core_components as dcc
import webviz_core_components as wcc

if TYPE_CHECKING:
    # pylint: disable=cyclic-import
    from ..property_statistics import PropertyStatistics


def ensemble_selector(
    parent: "PropertyStatistics", tab: str, multi: bool = False, value: str = None
) -> html.Div:
    return html.Div(
        style={"width": "75%"},
        children=[
            html.H5("Ensemble"),
            dcc.Dropdown(
                id={"id": parent.uuid("ensemble-selector"), "tab": tab},
                options=[
                    {"label": ens, "value": ens} for ens in parent.pmodel.ensembles
                ],
                multi=multi,
                value=value if value is not None else parent.pmodel.ensembles[0],
                clearable=False,
                persistence=True,
                persistence_type="session",
            ),
        ],
    )


def delta_ensemble_selector(
    parent: "PropertyStatistics", tab: str, multi: bool = False
) -> html.Div:
    return html.Div(
        style={"width": "75%"},
        children=[
            html.H5("Delta Ensemble"),
            dcc.Dropdown(
                id={"id": parent.uuid("delta-ensemble-selector"), "tab": tab},
                options=[
                    {"label": ens, "value": ens} for ens in parent.pmodel.ensembles
                ],
                multi=multi,
                value=parent.pmodel.ensembles[-1],
                clearable=False,
                persistence=True,
                persistence_type="session",
            ),
        ],
    )


def property_selector(
    parent: "PropertyStatistics", tab: str, multi: bool = False, value: str = None
) -> html.Div:
    display = "none" if len(parent.pmodel.properties) < 2 else "inline"
    return html.Div(
        style={"width": "75%", "display": display},
        children=[
            html.H5("Property"),
            dcc.Dropdown(
                id={"id": parent.uuid("property-selector"), "tab": tab},
                options=[
                    {"label": prop, "value": prop} for prop in parent.pmodel.properties
                ],
                multi=multi,
                value=value if value is not None else parent.pmodel.properties[0],
                clearable=False,
                persistence=True,
                persistence_type="session",
            ),
        ],
    )


def source_selector(
    parent: "PropertyStatistics", tab: str, multi: bool = False, value: str = None
) -> html.Div:
    display = "none" if len(parent.pmodel.sources) < 2 else "inline"
    return html.Div(
        style={"width": "75%", "display": display},
        children=[
            html.H5("Source"),
            dcc.Dropdown(
                id={"id": parent.uuid("source-selector"), "tab": tab},
                options=[
                    {"label": source, "value": source}
                    for source in parent.pmodel.sources
                ],
                multi=multi,
                value=value if value is not None else parent.pmodel.sources[0],
                clearable=False,
                persistence=True,
                persistence_type="session",
            ),
        ],
    )


def make_filter(
    parent: "PropertyStatistics",
    tab: str,
    df_column: str,
    column_values: list,
    multi: bool = True,
    value: Union[str, float] = None,
    open_details: bool = False,
) -> html.Div:
    return html.Div(
        children=html.Details(
            open=open_details,
            children=[
                html.Summary(df_column.lower().capitalize()),
                wcc.Select(
                    id={
                        "id": parent.uuid("filter-selector"),
                        "tab": tab,
                        "selector": df_column,
                    },
                    options=[{"label": i, "value": i} for i in column_values],
                    value=[value] if value is not None else column_values,
                    multi=multi,
                    size=min(20, len(column_values)),
                    persistence=True,
                    persistence_type="session",
                ),
            ],
        ),
    )


def filter_selector(
    parent: "PropertyStatistics",
    tab: str,
    multi: bool = True,
    value: Union[str, float] = None,
    open_details: bool = False,
) -> html.Div:
    return html.Div(
        children=[
            html.H5("Selections"),
            html.Div(
                children=[
                    make_filter(
                        parent=parent,
                        tab=tab,
                        df_column=sel,
                        column_values=parent.pmodel.selector_values(sel),
                        multi=multi,
                        value=value,
                        open_details=open_details,
                    )
                    for sel in parent.pmodel.selectors
                ]
            ),
        ]
    )



import xtgeo
import pandas as pd
import numpy as np
import plotly.graph_objects as go

from webviz_config.common_cache import CACHE


class HuvXsection:
    def __init__(
        self,
        surface_attributes=dict,
        zonation_status_file=None,
        well_points_file=None,
        zonelogname=None,
        well_attributes=dict,
    ):
        self.surface_attributes = surface_attributes
        self.zonation_status_file = zonation_status_file
        self.well_points_file = well_points_file
        self.zonelogname = zonelogname
        self.well_attributes = well_attributes
        self.planned_attributes = {}
        self.fig = None

    @CACHE.memoize(timeout=CACHE.TIMEOUT)
    def get_xsec_well_data(self, well_settings, well, is_planned=False):
        """Finds data for well to plot in cross section
        Args:
            well_settings: List of checked options from well settings modal button
            well: xtgeo well
            is_planned: boolean, if False well meta-data like zonelog is not added
        Returns:
            data: List with dictionary containing well, zonelog, zonation points
                  and conditional points
        """
        if well is None:
            return []
        data = [
            {
                "y": well.dataframe["Z_TVDSS"],
                "x": well.dataframe["R_HLEN"],
                "name": "well",
                "line": {"width": 7, "color": "black"},
                "fillcolor": "black",
            }
        ]
        if not is_planned:
            well_df = well.dataframe
            if "zonelog" in well_settings:
                data += self.get_zonelog_data(well, well_df, self.zonelogname)
            if "zonation_points" in well_settings:
                zonation_points = get_zonation_points(
                    well_df, well.wellname, self.zonation_status_file
                )
                data += [
                    {
                        "mode": "markers",
                        "y": zonation_points[1],
                        "x": zonation_points[0],
                        "name": "Zonation points",
                        "marker": {"size": 5, "color": "rgb(153,50,204)"},
                    }
                ]
            if "conditional_points" in well_settings:
                conditional_points = get_conditional_points(
                    well_df, well.wellname, self.well_points_file
                )
                data += [
                    {
                        "mode": "markers",
                        "y": conditional_points[1],
                        "x": conditional_points[0],
                        "name": "Conditional points",
                        "marker": {"size": 5, "color": "rgb(0,255,255)"},
                    }
                ]
        return data

    @CACHE.memoize(timeout=CACHE.TIMEOUT)
    def set_de_and_surface_lines(self, surfacefiles, de_keys, well, polyline):
        """Surface lines and corresponding depth error lines with fence from wellfile or polyline
        Args:
            surfacefiles: List of filepaths to surfacefiles
            de_keys: List of surfacepaths used as key in surface_attributes to access depth error
            well: xtgeo well
            polyline: Coordinates to polyline drawn in map view.
        """
        if well is None:
            fence = get_fencespec(polyline)
        else:
            fence = well.get_fence_polyline(nextend=100, sampling=5)
        for sfc_file in surfacefiles:
            sfc_line = self.surface_attributes[sfc_file]["surface"].get_randomline(
                fence
            )
            self.surface_attributes[sfc_file]["surface_line"] = sfc_line
            if sfc_file in de_keys:
                de_line = self.surface_attributes[sfc_file][
                    "surface_de"
                ].get_randomline(fence)
                sfc_line = self.surface_attributes[sfc_file]["surface_line"]
                self.surface_attributes[sfc_file]["de_line"] = de_line

    def get_xsec_layout(self, surfacefiles, well):
        """Scale cross section figure to well trajectory, well and surface intersection or polyline
        Args:
            surfacefiles: List of filepaths to surfacefiles
            well: xtgeo well
        Returns:
            layout: Dictionary with layout data
        """
        layout = {
            "yaxis": {
                "title": "True vertical depth [m]",
                "titlefont": {"size": 20},
                "tickfont": {"size": 16},
            },
            "xaxis": {
                "title": "Lateral length [m]",
                "titlefont": {"size": 18},
                "tickfont": {"size": 16},
            },
            "plot_bgcolor": "rgb(233,233,233)",
            "showlegend": False,
            "height": 830,
            "margin": {"t": 0, "l": 100},
        }
        if len(surfacefiles) == 0:
            layout["yaxis"].update({"autorange": "reversed"})

        elif well is None:
            ymin, ymax = self.sfc_lines_min_max_tvd(surfacefiles)
            layout["yaxis"].update({"range": [ymax, ymin]})

        else:
            y_min, y_max = self.sfc_lines_min_max_tvd(surfacefiles)
            x_min, x_max = get_range_from_well(well.dataframe, y_min)
            y_range = np.abs(y_max - y_min)
            x_range = np.abs(x_max - x_min)
            if y_range / x_range > 1:
                x_range = y_range + 10
            layout["yaxis"].update(
                {"range": [y_max + 0.15 * y_range, y_min - 0.15 * y_range]}
            )
            layout["xaxis"].update(
                {"range": [x_min - 0.35 * x_range, x_max + 0.35 * x_range]}
            )
        return layout

    def get_xsec_de_data(self, surfacefiles, de_keys):
        """Get cross section data for depth error
        Args:
            surfacefiles: List of filepaths to surfacefiles
            de_keys: List of surfacepaths used as key in surface_attributes to access depth error
        Returns:
            data: Dictionary with depth error data
        """
        common_files = [
            surfacefile_de
            for surfacefile_de in de_keys
            if surfacefile_de in surfacefiles
        ]
        data = []
        for sfc_file in common_files:
            data += [
                {
                    "x": self.surface_attributes[sfc_file]["surface_line"][:, 0],
                    "y": self.surface_attributes[sfc_file]["surface_line"][:, 1]
                    - self.surface_attributes[sfc_file]["de_line"][:, 1],
                    "line": {"color": "rgba(0,0,0,1)", "width": 0.6, "dash": "dash"},
                    "hoverinfo": "skip",
                    "mode": "lines",
                },
                {
                    "x": self.surface_attributes[sfc_file]["surface_line"][:, 0],
                    "y": self.surface_attributes[sfc_file]["surface_line"][:, 1]
                    + self.surface_attributes[sfc_file]["de_line"][:, 1],
                    "line": {"color": "rgba(0,0,0,1)", "width": 0.6, "dash": "dash"},
                    "fill": "tonexty",
                    "fillcolor": "rgba(0,0,0,0.2)",
                    "hoverinfo": "skip",
                    "mode": "lines",
                },
            ]
        return data

    def get_xsec_sfc_data(self, surfacefiles):
        """Get cross section data for surfaces
        Args:
            surfacefiles: List of filepaths to surfacefiles
        Returns:
            data: List containing dictionary with surface data or empty list if no surfacefiles
        """
        if len(surfacefiles) == 0:
            data = []
        else:
            _, _max = self.sfc_lines_min_max_tvd(surfacefiles)
            first_sfc_line = self.surface_attributes[surfacefiles[0]]["surface_line"]
            surface_tuples = [
                (sfc_file, self.surface_attributes[sfc_file]["order"])
                for sfc_file in surfacefiles
            ]
            surface_tuples.sort(key=stratigraphic_sort, reverse=True)

            data = [  # Create helpline for bottom of plot
                {
                    "x": [
                        first_sfc_line[0, 0],
                        first_sfc_line[np.shape(first_sfc_line)[0] - 1, 0],
                    ],
                    "y": [_max + 200, _max + 200],
                    "line": {"color": "rgba(0,0,0,1)", "width": 0.6},
                    "mode": "lines",
                }
            ]

            data += [
                {
                    "x": self.surface_attributes[sfc_file]["surface_line"][:, 0],
                    "y": self.surface_attributes[sfc_file]["surface_line"][:, 1],
                    "line": {"color": "rgba(0,0,0,1)", "width": 1},
                    "fill": "tonexty",
                    "fillcolor": self.surface_attributes[sfc_file]["color"],
                    "name": self.surface_attributes[sfc_file]["name"],
                    "text": self.get_hover_text(sfc_file),
                    "mode": "lines",
                    "hovertemplate": "<b>TVD:<b> %{y:.2f} <br>"
                    + "<b>TVD SD:<b> %{text}",
                }
                for sfc_file, _ in surface_tuples
            ]
        return data

    def sfc_lines_min_max_tvd(self, surfacefiles):
        """Find max and min TVD values of all surfaces
        Args:
            surfacefiles: List of filepaths to surfacefiles
        Returns:
            data: Numpy array containing max and min TVD values of all surfacefiles
        """
        maxvalues = np.array(
            [
                np.max(self.surface_attributes[sfc_file]["surface_line"][:, 1])
                for sfc_file in surfacefiles
            ]
        )
        minvalues = np.array(
            [
                np.min(self.surface_attributes[sfc_file]["surface_line"][:, 1])
                for sfc_file in surfacefiles
            ]
        )
        return np.min(minvalues), np.max(maxvalues)

    def get_hover_text(self, sfc_file):
        """Hover text for cross section graph to display depth error relative to surface line
        Args:
            sfc_file: Filepath to surfacefile
        Returns:
             Numpy array of TVD values from depth error surface line
        """
        return np.around(self.surface_attributes[sfc_file]["de_line"][:, 1], 2)

    @CACHE.memoize(timeout=CACHE.TIMEOUT)
    def set_xsec_fig(
        self, surfacefiles, de_keys, well_settings, well, is_planned=False
    ):
        """Set cross section plotly figure with data from wells, surfaces and depth error
        Args:
            surfacefiles: List of filepaths to surfacefiles
            de_keys: List of surfacepaths used as key in surface_attributes to access depth error
            well_settings: List of checked options from well settings modal button
            well: xtgeo well
            is_planned: boolean, if False well meta-data like zonelog is not added to plot
        """
        layout = self.get_xsec_layout(surfacefiles, well)
        data = (
            self.get_xsec_sfc_data(surfacefiles)
            + self.get_xsec_de_data(surfacefiles, de_keys)
            + self.get_xsec_well_data(well_settings, well, is_planned)
        )
        self.fig = go.Figure(dict({"data": data, "layout": layout}))

    @CACHE.memoize(timeout=CACHE.TIMEOUT)
    def get_intersection_dataframe(self, well):
        """Get intersection between surfaces and well with XTGeo
        Args:
            well: xtgeo well
        Returns:
            df: Dataframe with surfacename, TVD, depth uncertainty and direction
        """
        data = {
            "Surface name": [],
            "TVD [m]": [],
            "TVD SD [m]": [],
            "Direction": [],
        }
        for sfc_path in self.surface_attributes:
            sfc = self.surface_attributes[sfc_path]["surface"]
            err = self.surface_attributes[sfc_path]["surface_de"]
            with np.errstate(invalid="ignore"):
                surface_picks = well.get_surface_picks(sfc)
                # get_surface_picks raises warning when MD column is missing in well
            if surface_picks is not None:
                surface_picks_df = surface_picks.dataframe
                for _, row in surface_picks_df.iterrows():
                    surface_name = self.surface_attributes[sfc_path]["name"]
                    depth_uncertainty = err.get_value_from_xy(
                        point=(row["X_UTME"], row["Y_UTMN"])
                    )
                    data["Surface name"].append(surface_name)
                    data["TVD [m]"].append("%.2f" % row["Z_TVDSS"])
                    data["TVD SD [m]"].append("%.2f" % depth_uncertainty)
                    data["Direction"].append(row["DIRECTION"])
        return pd.DataFrame(data=data)

    @CACHE.memoize(timeout=CACHE.TIMEOUT)
    # pylint: disable=too-many-locals
    def get_zonelog_data(self, well, well_df, zonelogname="Zonelog"):
        """Find zonelogs where well trajectory intersects surfaces and assigns color.
        Args:
            well: XTGeo well from filepath to wellfile
            well_df: Dataframe of well
            zonelogname: Name of zonelog
        Returns:
            data: List containing dictionary with zonelog data
        """
        color_list = [None] * len(well.get_logrecord(zonelogname))
        for sfc_file in self.surface_attributes:
            for i, _ in enumerate(color_list):
                if well.get_logrecord_codename(zonelogname, i) == "dummy":
                    color_list[i] = "rgb(211,211,211)"
                if (
                    well.get_logrecord_codename(zonelogname, i)
                    == self.surface_attributes[sfc_file]["topofzone"]
                ):
                    self.surface_attributes[sfc_file]["zone_number"] = i
                    color_list[i] = self.surface_attributes[sfc_file]["color"]
        well_tvd = well_df["Z_TVDSS"].values.copy()
        well_rhlen = well_df["R_HLEN"].values.copy()
        zonevals = well_df[zonelogname].values
        zoneplot = []
        start = 0
        zone_transitions = np.where(zonevals[:-1] != zonevals[1:])
        for transition in zone_transitions:
            try:
                well_tvd = np.insert(well_tvd, transition, well_tvd[transition + 1])
                well_rhlen = np.insert(
                    well_rhlen, transition, well_rhlen[transition + 1]
                )
                zonevals = np.insert(zonevals, transition, zonevals[transition])
            except IndexError:
                pass
        for i in range(1, len(zonevals)):
            if (not np.isnan(zonevals[i])) and (np.isnan(zonevals[i - 1])):
                end = i - 1
                color = "rgb(211,211,211)"
                zoneplot.append(
                    {
                        "x": well_rhlen[start:end],
                        "y": well_tvd[start:end],
                        "line": {"width": 4, "color": color},
                        "name": f"Zone: {zonevals[i-1]}",
                    }
                )
                start = end + 1
            elif (np.isnan(zonevals[i])) and (not np.isnan(zonevals[i - 1])):
                end = i - 1
                color = color_list[int(zonevals[i - 1])]
                zoneplot.append(
                    {
                        "x": well_rhlen[start:end],
                        "y": well_tvd[start:end],
                        "line": {"width": 4, "color": color},
                        "name": f"Zone: {zonevals[i-1]}",
                    }
                )
                start = end + 1
            elif (
                (zonevals[i] != zonevals[i - 1])
                and (not np.isnan(zonevals[i]))
                and (not np.isnan(zonevals[i - 1]))
            ):
                end = i - 1
                color = color_list[int(zonevals[i - 1])]
                zoneplot.append(
                    {
                        "x": well_rhlen[start:end],
                        "y": well_tvd[start:end],
                        "line": {"width": 4, "color": color},
                        "name": f"Zone: {zonevals[i-1]}",
                    }
                )
                start = end + 1
        if not np.isnan(zonevals[-1]):
            end = len(zonevals) - 1
            color = color_list[int(zonevals[-1])]
            zoneplot.append(
                {
                    "x": well_rhlen[start:end],
                    "y": well_tvd[start:end],
                    "line": {"width": 4, "color": color},
                    "name": f"Zone: {zonevals[-2]}",
                }
            )
        else:
            color = "rgb(211,211,211)"
            zoneplot.append(
                {
                    "mode": "markers",
                    "x": np.array([well_rhlen[-2], well_rhlen[-1]]),
                    "y": np.array([well_tvd[-2], well_tvd[-1]]),
                    "marker": {"size": 3, "color": color},
                    "name": f"Zone: {zonevals[-1]}",
                }
            )
        return zoneplot


def stratigraphic_sort(elem):
    """ Sort surface lines in stratigraphic order"""
    return elem[1]


@CACHE.memoize(timeout=CACHE.TIMEOUT)
def get_zonation_points(well_df, wellname, zonation_status_file):
    """Finds zonation points along well trajectory
    Args:
        well_df: Dataframe of XTgeo well from filepath to wellfile
        wellname: Name of well
        zonation_status_file: Filepath to zonation_status.csv
    Returns:
        Numpy array with zone relative horizontal length and TVD
    """
    zonation_status_data = pd.read_csv(zonation_status_file)
    zone_df = zonation_status_data[zonation_status_data["Well"] == wellname]
    zone_df_xval = zone_df["x"].values.copy()
    zone_df_yval = zone_df["y"].values.copy()
    zone_rhlen = np.zeros(len(zone_df_xval))
    for i, _ in enumerate(zone_df_xval):
        well_df["XLEN"] = well_df["X_UTME"] - zone_df_xval[i]
        well_df["YLEN"] = well_df["Y_UTMN"] - zone_df_yval[i]
        well_df["SDIFF"] = np.sqrt(well_df.XLEN ** 2 + well_df.YLEN ** 2)
        index_array = np.where(well_df.SDIFF == well_df.SDIFF.min())
        zone_rhlen[i] = well_df["R_HLEN"].values[index_array[0]][0]
    return np.array([zone_rhlen, zone_df["TVD"]])


@CACHE.memoize(timeout=CACHE.TIMEOUT)
def get_conditional_points(well_df, wellname, well_points_file):
    """Finds conditional points where surfaces and well intersect
    Args:
        well_df: Dataframe with well from XTGeo
        wellname: Name of well
        well_points_file: Filepath to wellpoints.csv
    Returns:
        Numpy array of relative horizontal length and TVD
    """
    wellpoint_data = pd.read_csv(well_points_file)
    wellpoint_df = wellpoint_data[wellpoint_data["Well"] == wellname]
    wellpoint_df_xval = wellpoint_df["x"].values.copy()
    wellpoint_df_yval = wellpoint_df["y"].values.copy()
    cond_rhlen = np.zeros(len(wellpoint_df_xval))
    for i, _ in enumerate(wellpoint_df_xval):
        well_df["XLEN"] = well_df["X_UTME"] - wellpoint_df_xval[i]
        well_df["YLEN"] = well_df["Y_UTMN"] - wellpoint_df_yval[i]
        well_df["SDIFF"] = np.sqrt(well_df.XLEN ** 2 + well_df.YLEN ** 2)
        index_array = np.where(well_df.SDIFF == well_df.SDIFF.min())
        cond_rhlen[i] = well_df["R_HLEN"].values[index_array[0]][0]
    return np.array([cond_rhlen, wellpoint_df["TVD"]])


@CACHE.memoize(timeout=CACHE.TIMEOUT)
def get_range_from_well(well_df, ymin):
    """Finds min and max x values of well trajectory used in layout of cross section graph
    Args:
        well_df: Dataframe with well from XTgeo
    Returns:
        x_well_min: Float
        x_well_max: Float
    """
    x_well_max = np.max(well_df["R_HLEN"])
    x_well_min = 0
    for i in range(len(well_df["Z_TVDSS"])):
        if well_df["Z_TVDSS"][i] >= ymin:
            x_well_min = well_df["R_HLEN"][i]
            break
    return x_well_min, x_well_max


@CACHE.memoize(timeout=CACHE.TIMEOUT)
def get_fencespec(polyline):
    """Create a XTGeo fence spec from polyline coordinates
    Args:
        polyline: Polyline drawn in map view
    Returns:
        XTGeo fence of polyline
    """
    poly = xtgeo.Polygons()
    poly.dataframe = pd.DataFrame(
        [
            {
                "X_UTME": coordinates[1],
                "Y_UTMN": coordinates[0],
                "Z_TVDSS": 0,
                "POLY_ID": 1,
                "NAME": "polyline",
            }
            for coordinates in polyline
        ]
    )
    return poly.get_fence(asnumpy=True)



import dash_html_components as html
import dash_core_components as dcc

from .parameter_qc_view import parameter_qc_view
from .parameter_response_view import parameter_response_view


def main_view(parent) -> dcc.Tabs:
    tabs = [
        make_tab(
            label="Parameter distributions",
            children=parameter_qc_view(parent=parent),
        )
    ]
    if parent.vmodel is not None:
        tabs.append(
            make_tab(
                label="Parameters impact on simulation profiles",
                children=parameter_response_view(parent=parent),
            )
        )

    return html.Div(
        id=parent.uuid("layout"),
        children=dcc.Tabs(
            style={"width": "100%"},
            persistence=True,
            children=tabs,
        ),
    )


def make_tab(label, children):
    tab_style = {
        "borderBottom": "1px solid #d6d6d6",
        "padding": "6px",
        "fontWeight": "bold",
    }

    tab_selected_style = {
        "borderTop": "1px solid #d6d6d6",
        "borderBottom": "1px solid #d6d6d6",
        "backgroundColor": "#007079",
        "color": "white",
        "padding": "6px",
    }
    return dcc.Tab(
        label=label,
        style=tab_style,
        selected_style=tab_selected_style,
        children=children,
    )



from typing import List, Any, TYPE_CHECKING

import dash_html_components as html
import dash_core_components as dcc
import dash_table
import webviz_core_components as wcc
import webviz_subsurface_components as wsc

from .selector_view import (
    ensemble_selector,
    delta_ensemble_selector,
    property_selector,
    filter_selector,
    source_selector,
)

if TYPE_CHECKING:
    # pylint: disable=cyclic-import
    from ..property_statistics import PropertyStatistics


def surface_select_view(parent: "PropertyStatistics", tab: str) -> html.Div:
    return html.Div(
        id=parent.uuid("surface-select"),
        style={"width": "75%"},
        children=[
            html.Label("Surface statistics"),
            dcc.Dropdown(
                id={"id": parent.uuid("surface-type"), "tab": tab},
                options=[
                    {"label": "Mean", "value": "mean"},
                    {"label": "Standard Deviation", "value": "stddev"},
                    {"label": "Minimum", "value": "min"},
                    {"label": "Maximum", "value": "max"},
                    {"label": "P10", "value": "p10"},
                    {"label": "P90", "value": "p90"},
                ],
                clearable=False,
                value="mean",
                persistence=True,
                persistence_type="session",
            ),
        ],
    )


def surface_view(
    parent: "PropertyStatistics", ensemble: str, layers: list, synced_ids: list = None
) -> wsc.LeafletMap:
    return html.Div(
        style={"height": "22vh"},
        children=wsc.LeafletMap(
            id=f"{parent.uuid('surface-view-delta')}{ensemble}",
            layers=layers,
            syncedMaps=[
                f"{parent.uuid('surface-view-delta')}{s_id}" for s_id in synced_ids
            ]
            if synced_ids is not None
            else [],
            unitScale={},
            autoScaleMap=True,
            minZoom=-5,
            updateMode="replace",
            mouseCoords={"position": "bottomright"},
            colorBar={"position": "bottomleft"},
        ),
    )


# pylint: disable=too-many-arguments
def surface_views(
    parent: "PropertyStatistics",
    ensemble: str,
    delta_ensemble: str,
    ens_layer: dict,
    delta_ens_layer: dict,
    diff_layer: dict,
    prop: str,
    zone: str,
    statistic: str = "Mean",
) -> html.Div:
    return html.Div(
        style={"height": "76vh"},
        children=[
            html.P(
                style={"margin": "0px", "fontSize": "2vh"},
                children=f"{statistic} for {prop}, {zone} in {ensemble}",
            ),
            surface_view(
                parent=parent,
                ensemble=ensemble,
                synced_ids=[delta_ensemble, "diff"],
                layers=[ens_layer],
            ),
            html.P(
                style={"margin": "0px", "fontSize": "2vh"},
                children=f"{statistic} for {prop}, {zone} in {delta_ensemble}",
            ),
            surface_view(
                parent=parent,
                ensemble=delta_ensemble,
                synced_ids=[ensemble, "diff"],
                layers=[delta_ens_layer],
            ),
            html.P(
                style={"margin": "0px", "fontSize": "2vh"},
                children=f"{statistic} for {prop}, {zone} in {ensemble} - {delta_ensemble}",
            ),
            surface_view(
                parent=parent,
                ensemble="diff",
                synced_ids=[delta_ensemble, ensemble],
                layers=[diff_layer],
            ),
        ],
    )


def selector_view(parent: "PropertyStatistics") -> html.Div:
    return html.Div(
        style={"height": "80vh", "overflowY": "auto"},
        className="framed",
        children=[
            html.Div(
                children=[
                    ensemble_selector(parent=parent, tab="delta"),
                    delta_ensemble_selector(parent=parent, tab="delta"),
                ]
            ),
            property_selector(parent=parent, tab="delta"),
            source_selector(parent=parent, tab="delta"),
            filter_selector(parent=parent, tab="delta"),
        ],
    )


def delta_avg_view() -> html.Div:
    return html.Div(
        style={"flex": 1},
        children=[],
    )


def property_delta_view(parent: "PropertyStatistics") -> wcc.FlexBox:
    table_surf_options = [{"label": "Table view", "value": "table"}]
    if parent.surface_folders is not None:
        table_surf_options.append(
            {"label": "Surface view (click on bar in chart)", "value": "surface"}
        )
    return wcc.FlexBox(
        style={"margin": "20px"},
        children=[
            html.Div(style={"flex": 1}, children=selector_view(parent=parent)),
            html.Div(
                style={"flex": 4, "height": "80vh"},
                className="framed",
                children=[
                    dcc.RadioItems(
                        id=parent.uuid("delta-sort"),
                        options=[
                            {"label": "Sort by Average", "value": "Avg"},
                            {"label": "Sort by Standard Deviation", "value": "Stddev"},
                        ],
                        value="Avg",
                        labelStyle={
                            "display": "inline-block",
                            "margin": "5px",
                        },
                        persistence=True,
                        persistence_type="session",
                    ),
                    wcc.Graph(
                        id=parent.uuid("delta-bar-graph"),
                        config={"displayModeBar": False},
                        style={"height": "75vh"},
                    ),
                ],
            ),
            html.Div(
                style={"flex": 4, "height": "80vh"},
                className="framed",
                children=[
                    dcc.RadioItems(
                        id=parent.uuid("delta-switch-table-surface"),
                        options=table_surf_options,
                        value="table",
                        labelStyle={
                            "display": "inline-block",
                            "margin": "5px",
                        },
                        persistence=True,
                        persistence_type="session",
                    ),
                    html.Div(id=parent.uuid("delta-table-surface-wrapper")),
                ],
            ),
        ],
    )


def table_view(data: List[Any], columns: List[Any]) -> html.Div:
    return html.Div(
        style={"fontSize": "1rem"},
        children=dash_table.DataTable(
            sort_action="native",
            page_action="native",
            filter_action="native",
            style_table={
                "height": "74vh",
                "overflow": "auto",
            },
            data=data,
            columns=columns,
        ),
    )



from typing import Tuple, Union
from itertools import chain

import numpy as np
import pandas as pd
from dash.dependencies import Input, Output, State, ALL
from dash.exceptions import PreventUpdate
import dash
import dash_core_components as dcc
import plotly.express as px
import plotly.graph_objects as go

from ..utils.colors import find_intermediate_color
from ..figures.correlation_figure import CorrelationFigure
from ..utils.colors import hex_to_rgb, rgb_to_hex


# pylint: disable=too-many-statements,
def parameter_response_controller(parent, app):
    @app.callback(
        Output(parent.uuid("vector-vs-time-graph"), "figure"),
        Output(parent.uuid("vector-vs-param-scatter"), "figure"),
        Output(parent.uuid("vector-corr-graph"), "figure"),
        Output(parent.uuid("param-corr-graph"), "figure"),
        Input({"id": parent.uuid("ensemble-selector"), "tab": "response"}, "value"),
        Input(parent.uuid("vector-select"), "children"),
        Input({"id": parent.uuid("parameter-select"), "tab": "response"}, "value"),
        Input(parent.uuid("date-selected"), "children"),
        Input({"id": parent.uuid("vtype-filter"), "tab": "response"}, "value"),
        Input(
            {
                "id": parent.uuid("vitem-filter"),
                "tab": "response",
                "vtype": ALL,
            },
            "value",
        ),
        Input({"id": parent.uuid("plot-options"), "tab": "response"}, "value"),
        State(parent.uuid("vector-vs-time-graph"), "figure"),
        State(parent.uuid("param-corr-graph"), "figure"),
        State(parent.uuid("vector-corr-graph"), "figure"),
        State(parent.uuid("vector-vs-param-scatter"), "figure"),
    )
    # pylint: disable=too-many-locals, too-many-arguments
    def _update_graphs(
        ensemble: str,
        vector: str,
        parameter: Union[None, dict],
        date: str,
        vector_type_filter: list,
        vector_item_filters: list,
        options: str,
        timeseries_fig: dict,
        corr_p_fig: dict,
        corr_v_fig: dict,
        scatter_fig: dict,
    ) -> Tuple[dict, dict, dict, dict]:
        """
        Main callback to update plots. Initially all plots are generated,
        while only relevant plots are updated in subsequent callbacks
        """

        if (
            dash.callback_context.triggered is None
            or dash.callback_context.triggered[0]["prop_id"] == "."
            or vector is None
        ):
            raise PreventUpdate
        ctx = dash.callback_context.triggered[0]["prop_id"].split(".")[0]

        initial_run = timeseries_fig is None
        color = options["color"] if options["color"] is not None else "#007079"
        daterange = parent.vmodel.daterange_for_plot(vector=vector)

        # Make timeseries graph
        if relevant_ctx(parent, ctx, operation="timeseries_fig") or initial_run:
            timeseries_fig = update_timeseries_graph(
                parent.vmodel,
                ensemble,
                vector,
                xaxisrange=[min(daterange[0], date), max(daterange[1], date)],
                real_filter=None,
            )

        if parent.uuid("plot-options") not in ctx or initial_run:
            vectors_filtered = filter_vectors(
                parent, vector_type_filter, vector_item_filters
            )
            if vector not in vectors_filtered:
                vectors_filtered.append(vector)
            merged_df = merge_parameter_and_vector_df(
                parent, ensemble, vectors_filtered, date
            )

        # Make correlation figure for vector
        if options["autocompute_corr"] and (
            relevant_ctx(parent, ctx, operation="vector_correlation") or initial_run
        ):
            corr_v_fig = make_correlation_figure(
                merged_df, response=vector, corrwith=parent.pmodel.parameters
            ).figure

        # Get clicked parameter correlation bar or largest bar initially
        parameter = (
            parameter if parameter is not None else corr_v_fig["data"][0]["y"][-1]
        )
        corr_v_fig = color_corr_bars(corr_v_fig, parameter, color, options["opacity"])

        # Make correlation figure for parameter
        if options["autocompute_corr"] and (
            relevant_ctx(parent, ctx, operation="parameter_correlation") or initial_run
        ):
            corr_p_fig = make_correlation_figure(
                merged_df, response=parameter, corrwith=vectors_filtered
            ).figure

        corr_p_fig = color_corr_bars(corr_p_fig, vector, color, options["opacity"])

        # Create scatter plot of vector vs parameter
        if relevant_ctx(parent, ctx, operation="scatter") or initial_run:
            scatter_fig = update_scatter_graph(merged_df, vector, parameter, color)

        scatter_fig = scatter_fig_color_update(scatter_fig, color, options["opacity"])

        # Order realizations sorted on value of parameter and color traces
        df_value_norm = parent.pmodel.get_real_and_value_df(
            ensemble, parameter=parameter, normalize=True
        )
        if relevant_ctx(parent, ctx, operation="color_timeseries_fig") or initial_run:
            timeseries_fig = color_timeseries_graph(
                timeseries_fig, ensemble, parameter, vector, df_value_norm
            )

        # Draw date selected as line
        timeseries_fig = add_date_line(timeseries_fig, date, options["show_dateline"])

        # Ensure xaxis covers selected date
        if parent.uuid("date-selected") in ctx:
            timeseries_fig["layout"]["xaxis"].update(
                range=[min(daterange[0], date), max(daterange[1], date)]
            )

        return timeseries_fig, scatter_fig, corr_v_fig, corr_p_fig

    @app.callback(
        Output(parent.uuid("date-slider"), "value"),
        Input(parent.uuid("vector-vs-time-graph"), "clickData"),
    )
    def _update_date_from_clickdata(timeseries_clickdata: Union[None, dict]):
        """Update date-slider from clickdata"""
        dates = parent.vmodel.dates
        return (
            dates.index(timeseries_clickdata.get("points", [{}])[0]["x"])
            if timeseries_clickdata is not None
            else len(dates) - 1
        )

    @app.callback(
        Output(parent.uuid("date-selected"), "children"),
        Input(parent.uuid("date-slider"), "value"),
    )
    def _update_date(dateidx: int):
        """Update selected date from date-slider"""
        return parent.vmodel.dates[dateidx]

    @app.callback(
        Output({"id": parent.uuid("plot-options"), "tab": "response"}, "value"),
        Input({"id": parent.uuid("checkbox-options"), "tab": "response"}, "value"),
        Input({"id": parent.uuid("color-selector"), "tab": "response"}, "clickData"),
        Input({"id": parent.uuid("opacity-selector"), "tab": "response"}, "value"),
    )
    def _update_plot_options(
        checkbox_options: list,
        color_clickdata: str,
        opacity: float,
    ):
        """Combine plot options in one dictionary"""
        ctx = dash.callback_context.triggered[0]["prop_id"].split(".")[0]
        if color_clickdata is not None:
            color = color_clickdata["points"][0]["marker.color"]
            if "rgb" in color:
                color = rgb_to_hex(color)

        return dict(
            show_dateline="DateLine" in checkbox_options,
            autocompute_corr="AutoCompute" in checkbox_options,
            color=None if color_clickdata is None else color,
            opacity=opacity,
            ctx=ctx,
        )

    @app.callback(
        Output(parent.uuid("vector-select"), "children"),
        Input(parent.uuid("vshort-select"), "value"),
        Input({"id": parent.uuid("vitem-select"), "shortname": ALL}, "value"),
    )
    def _combine_substrings_to_vector(shortname: str, item: list):
        """Combine vector shortname and item to full vector name"""
        vector = shortname if not item or item[0] is None else f"{shortname}:{item[0]}"

        if vector not in parent.vmodel.vectors:
            raise PreventUpdate
        return vector

    @app.callback(
        Output(parent.uuid("vshort-select"), "options"),
        Output(parent.uuid("vshort-select"), "value"),
        Output(parent.uuid("clickdata-store"), "data"),
        Input({"id": parent.uuid("vtype-select"), "state": ALL}, "value"),
        Input(parent.uuid("param-corr-graph"), "clickData"),
    )
    def _update_vectorlist(vtype: list, corr_param_clickdata: dict):
        """
        Update the vector shortname options and selected value from
        selected vector type or clickdata
        """
        ctx = dash.callback_context.triggered[0]["prop_id"].split(".")[0]
        click_data = parent.uuid("param-corr-graph") in ctx

        vtype = vtype[0]

        if click_data:
            vector_selected = corr_param_clickdata.get("points", [{}])[0].get("y")
            vtype = find_vector_type(parent, vector_selected)

        shortname = (
            vector_selected.split(":")[0]
            if click_data
            else parent.vmodel.vector_groups[vtype]["shortnames"][0]
        )

        return (
            [
                {"label": i, "value": i}
                for i in parent.vmodel.vector_groups[vtype]["shortnames"]
            ],
            shortname,
            dict(
                vector=vector_selected,
                shortname=shortname,
                vtype=vtype,
            )
            if click_data
            else {},
        )

    @app.callback(
        Output(parent.uuid("vitems-container"), "children"),
        Output(parent.uuid("vtype-container"), "children"),
        Input(parent.uuid("vshort-select"), "value"),
        State({"id": parent.uuid("vtype-select"), "state": ALL}, "value"),
        State(parent.uuid("clickdata-store"), "data"),
        State({"id": parent.uuid("vitem-select"), "shortname": ALL}, "value"),
    )
    def _update_vector_items(
        shortname: str,
        previous_vtype: list,
        clickdata_vector: dict,
        previous_item: list,
    ):
        """
        Update the div container for vector type and vector item selections
        from clickdata and selected vector shortname
        """
        if clickdata_vector and shortname != clickdata_vector["shortname"]:
            clickdata_vector = {}
        vtype = clickdata_vector["vtype"] if clickdata_vector else previous_vtype[0]

        items = [
            v
            for v in parent.vmodel.vector_groups[vtype]["items"]
            if f"{shortname}:{v}" in parent.vmodel.vectors
        ]
        if items and not clickdata_vector:
            item = previous_item[0] if previous_item[0] in items else items[0]
        if items and clickdata_vector:
            item = clickdata_vector["vector"].replace(f"{shortname}:", "")

        return (
            [
                dcc.Dropdown(
                    id={"id": parent.uuid("vitem-select"), "shortname": shortname},
                    options=[{"label": i, "value": i} for i in items],
                    value=item if items else None,
                    disabled=not items,
                    placeholder="No subselections...",
                    clearable=False,
                    persistence=True,
                    persistence_type="session",
                )
            ],
            [
                dcc.RadioItems(
                    id={"id": parent.uuid("vtype-select"), "state": "update"},
                    options=[
                        {"label": i, "value": i} for i in parent.vmodel.vector_groups
                    ],
                    value=vtype,
                    labelStyle={"display": "inline-block", "margin-right": "10px"},
                )
            ]
            if clickdata_vector and clickdata_vector["vtype"] != previous_vtype[0]
            else dash.no_update,
        )

    @app.callback(
        Output({"id": parent.uuid("parameter-select"), "tab": "response"}, "value"),
        Input(parent.uuid("vector-corr-graph"), "clickData"),
    )
    def _update_parameter_selected(
        corr_vector_clickdata: Union[None, dict],
    ) -> str:
        """Update the selected parameter from clickdata"""
        if corr_vector_clickdata is None:
            raise PreventUpdate
        return corr_vector_clickdata.get("points", [{}])[0].get("y")


# pylint: disable=inconsistent-return-statements
def relevant_ctx(parent, ctx: list, operation: str):
    """Group relevant uuids for the different plots"""
    vector = parent.uuid("vector-select") in ctx
    date = parent.uuid("date-selected") in ctx
    parameter = parent.uuid("parameter-select") in ctx
    ensemble = parent.uuid("ensemble-selector") in ctx
    filtered_vectors = (
        parent.uuid("vtype-filter") in ctx or parent.uuid("vitem-filter") in ctx
    )

    if operation == "timeseries_fig":
        return any([vector, ensemble])
    if operation == "scatter":
        return any([vector, date, parameter, ensemble])
    if operation == "parameter_correlation":
        return any([filtered_vectors, date, parameter, ensemble])
    if operation == "vector_correlation":
        return any([vector, date, ensemble])
    if operation == "color_timeseries_fig":
        return any([parameter, ensemble, vector])


def find_vector_type(parent, vector: str):
    """Get vector type from vector"""
    for vgroup, values in parent.vmodel.vector_groups.items():
        if vector in values["vectors"]:
            return vgroup
    return None


def filter_vectors(parent, vector_types: list, vector_items: list):
    """Filter vector list used for correlation"""
    vectors = list(
        chain.from_iterable(
            [parent.vmodel.vector_groups[vtype]["vectors"] for vtype in vector_types]
        )
    )
    items = list(chain.from_iterable(vector_items))
    filtered_vectors_with_items = [
        v for v in vectors if any(v.split(":")[1] == x for x in items if ":" in v)
    ]
    return [v for v in vectors if v in filtered_vectors_with_items or ":" not in v]


def update_timeseries_graph(
    timeseries_model,
    ensemble: str,
    vector: str,
    xaxisrange: list,
    real_filter: pd.Series = None,
):
    """Create vector vs time plot"""
    return {
        "data": timeseries_model.add_realization_traces(
            ensemble=ensemble, vector=vector, real_filter=real_filter
        ),
        "layout": dict(
            margin={"r": 40, "l": 20, "t": 60, "b": 20},
            yaxis={"automargin": True},
            xaxis={"range": xaxisrange},
            hovermode="closest",
            paper_bgcolor="white",
            plot_bgcolor="white",
            showlegend=False,
        ),
    }


def add_date_line(figure: dict, selected_date: str, show_dateline: bool):
    """Add/remove dateline on timeseries graph."""
    dateline_idx = [
        idx for idx, trace in enumerate(figure["data"]) if trace["name"] == "Dateline"
    ]
    if dateline_idx and show_dateline:
        figure["data"][dateline_idx[0]].update(
            x=[selected_date, selected_date], text=["", selected_date]
        )
    if dateline_idx and not show_dateline:
        figure["data"].pop(int(dateline_idx[0]))

    if not dateline_idx and show_dateline:
        ymin = min([min(trace["y"]) for trace in figure["data"]])
        ymax = max([max(trace["y"]) for trace in figure["data"]])
        figure["data"].append(
            go.Scatter(
                x=[selected_date, selected_date],
                y=[ymin, ymax],
                cliponaxis=False,
                mode="lines+text",
                line={"dash": "dot", "width": 4, "color": "#243746"},
                name="Dateline",
                text=["", selected_date],
                textposition="top center",
            )
        )
    return figure


def color_timeseries_graph(
    figure: dict,
    ensemble: str,
    selected_param: str,
    vector: str,
    df_norm: pd.DataFrame = None,
):
    """Color timeseries lines by parameter value"""
    if df_norm is not None:
        for trace in figure.get("data", []):
            if trace["name"] == ensemble:
                trace["marker"]["color"] = set_real_color(
                    real_no=trace["customdata"], df_norm=df_norm
                )
                trace["hovertext"] = (
                    f"Real: {str(trace['customdata'])}, {selected_param}: "
                    f"{df_norm.loc[df_norm['REAL'] == trace['customdata']].iloc[0]['VALUE']}"
                )
        figure["layout"]["title"] = {
            "text": f"{vector} colored by {selected_param}",
        }

    return figure


def set_real_color(df_norm, real_no: str):
    """
    Return color for trace based on normalized parameter value.
    Midpoint for the colorscale is set on the average value
    """
    red = "rgba(255,18,67, 1)"
    mid_color = "rgba(220,220,220,1)"
    green = "rgba(62,208,62, 1)"

    mean = df_norm["VALUE_NORM"].mean()

    norm_value = df_norm.loc[df_norm["REAL"] == real_no].iloc[0]["VALUE_NORM"]
    if norm_value <= mean:
        intermed = norm_value / mean
        return find_intermediate_color(red, mid_color, intermed, colortype="rgba")
    if norm_value > mean:
        intermed = (norm_value - mean) / (1 - mean)
        return find_intermediate_color(mid_color, green, intermed, colortype="rgba")
    return "rgba(220,220,220, 0.2)"


def merge_parameter_and_vector_df(parent, ensemble: str, vectors: list, date: str):
    """Merge parameter dataframe with vector dataframe on given date """
    # Get dataframe with vector and REAL
    vector_df = parent.vmodel.get_ensemble_vectors_for_date(
        ensemble=ensemble,
        vectors=vectors,
        date=date,
    ).copy()
    vector_df["REAL"] = vector_df["REAL"].astype(int)
    # Get dataframe with parameters
    param_df = parent.pmodel.dataframe.copy()
    param_df = param_df[param_df["ENSEMBLE"] == ensemble]
    param_df["REAL"] = param_df["REAL"].astype(int)
    # Return merged dataframe
    param_df.set_index("REAL", inplace=True)
    vector_df.set_index("REAL", inplace=True)
    return vector_df.join(param_df).reset_index()


def update_scatter_graph(
    df: pd.DataFrame, vector: str, selected_param: str, color: str = None
):
    """Create scatter plot of selected vector vs selected parameter"""
    return (
        px.scatter(
            df[[vector, selected_param]],
            x=selected_param,
            y=vector,
            trendline="ols" if df[vector].nunique() > 1 else None,
            trendline_color_override="#243746",
        )
        .update_layout(
            margin={
                "r": 20,
                "l": 20,
                "t": 60,
                "b": 20,
            },
            paper_bgcolor="white",
            plot_bgcolor="white",
            title={"text": f"{vector} vs {selected_param}", "x": 0.5},
            xaxis_title=None,
            yaxis_title=None,
        )
        .update_traces(
            marker={
                "size": 15,
                "color": hex_to_rgb(color, 0.7),
                "line": {"width": 1.2, "color": hex_to_rgb(color, 1)},
            }
        )
    )


def scatter_fig_color_update(figure: dict, color: str, opacity: float):
    """Update color for scatter plot"""
    for trace in figure["data"]:
        if trace["mode"] == "markers":
            trace["marker"].update(color=hex_to_rgb(color, opacity))
            trace["marker"]["line"].update(color=hex_to_rgb(color, 1))
    return figure


def make_correlation_figure(df: pd.DataFrame, response: str, corrwith: list):
    """Create a bar plot with correlations for chosen response"""
    corrseries = correlate(df[corrwith + [response]], response=response)
    return CorrelationFigure(
        corrseries, n_rows=15, title=f"Correlations with {response}"
    )


def correlate(df: pd.DataFrame, response: str):
    """Returns the correlation matrix for a dataframe"""
    df = df[df.columns[df.nunique() > 1]].copy()
    if response not in df.columns:
        df[response] = np.nan
    series = df[response]
    df = df.drop(columns=[response])
    corrdf = df.corrwith(series)
    return corrdf.reindex(corrdf.abs().sort_values().index)


def color_corr_bars(
    figure: dict,
    selected_bar: str,
    color: str,
    opacity: float,
    color_selected="#FF1243",
):
    """
    Set colors to the correlation plot bar,
    with separate color for the selected bar
    """
    figure["data"][0]["marker"] = {
        "color": [
            hex_to_rgb(color, opacity)
            if _bar != selected_bar
            else hex_to_rgb(color_selected, 0.8)
            for _bar in figure["data"][0]["y"]
        ],
        "line": {
            "color": [
                color if _bar != selected_bar else color_selected
                for _bar in figure["data"][0]["y"]
            ],
            "width": 1.2,
        },
    }
    return figure



from uuid import uuid4
from pathlib import Path

import numpy as np
import pandas as pd
from plotly.subplots import make_subplots
from dash.dependencies import Input, Output
import dash_html_components as html
import dash_core_components as dcc
import webviz_core_components as wcc
from webviz_config.webviz_store import webvizstore
from webviz_config.common_cache import CACHE
from webviz_config import WebvizPluginABC
from webviz_config.utils import calculate_slider_step

from webviz_subsurface._models import EnsembleSetModel
from .._datainput.fmu_input import load_parameters, load_csv


class ParameterResponseCorrelation(WebvizPluginABC):
    """Visualizes correlations between numerical input parameters and responses.

---
**Three main options for input data: Aggregated, file per realization and read from UNSMRY.**

**Using aggregated data**
* **`parameter_csv`:** Aggregated csvfile for input parameters with `REAL` and `ENSEMBLE` columns \
(absolute path or relative to config file).
* **`response_csv`:** Aggregated csvfile for response parameters with `REAL` and `ENSEMBLE` \
columns (absolute path or relative to config file).


**Using a response file per realization**
* **`ensembles`:** Which ensembles in `shared_settings` to visualize.
* **`response_file`:** Local (per realization) csv file for response parameters (Cannot be \
                    combined with `response_csv` and `parameter_csv`).


**Using simulation time series data directly from `UNSMRY` files as responses**
* **`ensembles`:** Which ensembles in `shared_settings` to visualize. The lack of `response_file` \
                implies that the input data should be time series data from simulation `.UNSMRY` \
                files, read using `fmu-ensemble`.
* **`column_keys`:** (Optional) slist of simulation vectors to include as responses when reading \
                from UNSMRY-files in the defined ensembles (default is all vectors). * can be \
                used as wild card.
* **`sampling`:** (Optional) sampling frequency when reading simulation data directly from \
               `.UNSMRY`-files (default is monthly).

?> The `UNSMRY` input method implies that the "DATE" vector will be used as a filter \
   of type `single` (as defined below under `response_filters`).


**Common settings for all input options**

All of these are optional, some have defaults seen in the code snippet below.

* **`response_filters`:** Optional dictionary of responses (columns in csv file or simulation \
                       vectors) that can be used as row filtering before aggregation. \
                       Valid options:
    * `single`: Dropdown with single selection.
    * `multi`: Dropdown with multiple selection.
    * `range`: Slider with range selection.
* **`response_ignore`:** List of response (columns in csv or simulation vectors) to ignore \
                      (cannot use with response_include).
* **`response_include`:** List of response (columns in csv or simulation vectors) to include \
                       (cannot use with response_ignore).
* **`aggregation`:** How to aggregate responses per realization. Either `sum` or `mean`.
* **`corr_method`:** Correlation method. Either `pearson` or `spearman`.

---

?> Non-numerical (string-based) input parameters and responses are removed.

?> The responses will be aggregated per realization; meaning that if your filters do not reduce \
the response to a single value per realization in your data, the values will be aggregated \
accoording to your defined `aggregation`. If e.g. the response is a form of volume, \
and the filters are regions (or other subdivisions of the total volume), then `sum` would \
be a natural aggregation. If on the other hand the response is the pressures in the \
same volume, aggregation as `mean` over the subdivisions of the same volume \
would make more sense (though the pressures in this case would not be volume weighted means, \
and the aggregation would therefore likely be imprecise).

!> It is **strongly recommended** to keep the data frequency to a regular frequency (like \
`monthly` or `yearly`). This applies to both csv input and when reading from `UNSMRY` \
(controlled by the `sampling` key). This is because the statistics are calculated per DATE over \
all realizations in an ensemble, and the available dates should therefore not differ between \
individual realizations of an ensemble.

**Using aggregated data**

The `parameter_csv` file must have columns `REAL`, `ENSEMBLE` and the parameter columns.

The `response_csv` file must have columns `REAL`, `ENSEMBLE` and the response columns \
(and the columns to use as `response_filters`, if that option is used).


**Using a response file per realization**

Parameters are extracted automatically from the `parameters.txt` files in the individual
realizations, using the `fmu-ensemble` library.

The `response_file` must have the response columns (and the columns to use as `response_filters`, \
if that option is used).


**Using simulation time series data directly from `UNSMRY` files as responses**

Parameters are extracted automatically from the `parameters.txt` files in the individual
realizations, using the `fmu-ensemble` library.

Responses are extracted automatically from the `UNSMRY` files in the individual realizations,
using the `fmu-ensemble` library.

!> The `UNSMRY` files are auto-detected by `fmu-ensemble` in the `eclipse/model` folder of the \
individual realizations. You should therefore not have more than one `UNSMRY` file in this \
folder, to avoid risk of not extracting the right data.
"""

    # pylint:disable=too-many-arguments
    def __init__(
        self,
        app,
        parameter_csv: Path = None,
        response_csv: Path = None,
        ensembles: list = None,
        response_file: str = None,
        response_filters: dict = None,
        response_ignore: list = None,
        response_include: list = None,
        column_keys: list = None,
        sampling: str = "monthly",
        aggregation: str = "sum",
        corr_method: str = "pearson",
    ):

        super().__init__()

        self.parameter_csv = parameter_csv if parameter_csv else None
        self.response_csv = response_csv if response_csv else None
        self.response_file = response_file if response_file else None
        self.response_filters = response_filters if response_filters else {}
        self.response_ignore = response_ignore if response_ignore else None
        self.column_keys = column_keys
        self.time_index = sampling
        self.corr_method = corr_method
        self.aggregation = aggregation
        if response_ignore and response_include:
            raise ValueError(
                'Incorrent argument. either provide "response_include", '
                '"response_ignore" or neither'
            )
        if parameter_csv and response_csv:
            if ensembles or response_file:
                raise ValueError(
                    'Incorrect arguments. Either provide "csv files" or '
                    '"ensembles and response_file".'
                )
            self.parameterdf = read_csv(self.parameter_csv)
            self.responsedf = read_csv(self.response_csv)

        elif ensembles:
            self.ens_paths = {
                ens: app.webviz_settings["shared_settings"]["scratch_ensembles"][ens]
                for ens in ensembles
            }
            self.parameterdf = load_parameters(
                ensemble_paths=self.ens_paths, ensemble_set_name="EnsembleSet"
            )
            if self.response_file:
                self.responsedf = load_csv(
                    ensemble_paths=self.ens_paths,
                    csv_file=response_file,
                    ensemble_set_name="EnsembleSet",
                )
            else:
                self.emodel = EnsembleSetModel(ensemble_paths=self.ens_paths)
                self.responsedf = self.emodel.load_smry(
                    column_keys=self.column_keys,
                    time_index=self.time_index,
                )
                self.response_filters["DATE"] = "single"
        else:
            raise ValueError(
                'Incorrect arguments. Either provide "csv files" or "ensembles and response_file".'
            )
        self.check_runs()
        self.check_response_filters()
        if response_ignore:
            self.responsedf.drop(response_ignore, errors="ignore", axis=1, inplace=True)
        if response_include:
            self.responsedf.drop(
                self.responsedf.columns.difference(
                    [
                        "REAL",
                        "ENSEMBLE",
                        *response_include,
                        *list(response_filters.keys()),
                    ]
                ),
                errors="ignore",
                axis=1,
                inplace=True,
            )

        self.plotly_theme = app.webviz_settings["theme"].plotly_theme
        self.uid = uuid4()
        self.set_callbacks(app)

    def ids(self, element):
        """Generate unique id for dom element"""
        return f"{element}-id-{self.uid}"

    @property
    def tour_steps(self):
        steps = [
            {
                "id": self.ids("layout"),
                "content": (
                    "Dashboard displaying correlation between selected "
                    "response and input parameters."
                ),
            },
            {
                "id": self.ids("correlation-graph"),
                "content": (
                    "Visualization of the correlations between currently selected "
                    "response and input parameters ranked by the absolute correlation "
                    "coefficient. Click on any correlation to visualize the distribution "
                    "between that parameter and the response."
                ),
            },
            {
                "id": self.ids("distribution-graph"),
                "content": (
                    "Visualized the distribution of the response and the selected input parameter "
                    "in the correlation chart."
                ),
            },
            {
                "id": self.ids("ensemble"),
                "content": ("Select the active ensemble."),
            },
            {
                "id": self.ids("responses"),
                "content": ("Select the active response."),
            },
        ]

        return steps

    @property
    def responses(self):
        """Returns valid responses. Filters out non numerical columns,
        and filterable columns"""
        responses = list(
            self.responsedf.drop(["ENSEMBLE", "REAL"], axis=1)
            .apply(pd.to_numeric, errors="coerce")
            .dropna(how="all", axis="columns")
            .columns
        )
        return [p for p in responses if p not in self.response_filters.keys()]

    @property
    def parameters(self):
        """Returns numerical input parameters"""
        parameters = list(
            self.parameterdf.drop(["ENSEMBLE", "REAL"], axis=1)
            .apply(pd.to_numeric, errors="coerce")
            .dropna(how="all", axis="columns")
            .columns
        )
        return parameters

    @property
    def ensembles(self):
        """Returns list of ensembles"""
        return list(self.parameterdf["ENSEMBLE"].unique())

    def check_runs(self):
        """Check that input parameters and response files have
        the same number of runs"""
        for col in ["ENSEMBLE", "REAL"]:
            if sorted(list(self.parameterdf[col].unique())) != sorted(
                list(self.responsedf[col].unique())
            ):
                raise ValueError("Parameter and response files have different runs")

    def check_response_filters(self):
        """'Check that provided response filters are valid"""
        if self.response_filters:
            for col_name, col_type in self.response_filters.items():
                if col_name not in self.responsedf.columns:
                    raise ValueError(f"{col_name} is not in response file")
                if col_type not in ["single", "multi", "range"]:
                    raise ValueError(
                        f"Filter type {col_type} for {col_name} is not valid."
                    )

    @property
    def filter_layout(self):
        """Layout to display selectors for response filters"""
        children = []
        for col_name, col_type in self.response_filters.items():
            domid = self.ids(f"filter-{col_name}")
            values = list(self.responsedf[col_name].unique())
            if col_type == "multi":
                selector = wcc.Select(
                    id=domid,
                    options=[{"label": val, "value": val} for val in values],
                    value=values,
                    multi=True,
                    size=min(20, len(values)),
                    persistence=True,
                    persistence_type="session",
                )
            elif col_type == "single":
                selector = dcc.Dropdown(
                    id=domid,
                    options=[{"label": val, "value": val} for val in values],
                    value=values[0],
                    multi=False,
                    clearable=False,
                    persistence=True,
                    persistence_type="session",
                )
            elif col_type == "range":
                selector = make_range_slider(domid, self.responsedf[col_name], col_name)
            else:
                return children
            children.append(
                html.Div(
                    children=[
                        html.Label(col_name),
                        selector,
                    ]
                )
            )

        return children

    @property
    def control_layout(self):
        """Layout to select e.g. iteration and response"""
        return [
            html.Div(
                [
                    html.Label("Ensemble"),
                    dcc.Dropdown(
                        id=self.ids("ensemble"),
                        options=[
                            {"label": ens, "value": ens} for ens in self.ensembles
                        ],
                        clearable=False,
                        value=self.ensembles[0],
                        persistence=True,
                        persistence_type="session",
                    ),
                ]
            ),
            html.Div(
                [
                    html.Label("Response"),
                    dcc.Dropdown(
                        id=self.ids("responses"),
                        options=[
                            {"label": ens, "value": ens} for ens in self.responses
                        ],
                        clearable=False,
                        value=self.responses[0],
                        persistence=True,
                        persistence_type="session",
                    ),
                ]
            ),
        ]

    @property
    def layout(self):
        """Main layout"""
        return wcc.FlexBox(
            id=self.ids("layout"),
            children=[
                html.Div(
                    style={"flex": 3},
                    children=[
                        wcc.Graph(self.ids("correlation-graph")),
                        dcc.Store(
                            id=self.ids("initial-parameter"), storage_type="session"
                        ),
                    ],
                ),
                html.Div(
                    style={"flex": 3},
                    children=wcc.Graph(self.ids("distribution-graph")),
                ),
                html.Div(
                    style={"flex": 1},
                    children=self.control_layout + self.filter_layout
                    if self.response_filters
                    else [],
                ),
            ],
        )

    @property
    def correlation_input_callbacks(self):
        """List of Inputs for correlation callback"""
        callbacks = [
            Input(self.ids("ensemble"), "value"),
            Input(self.ids("responses"), "value"),
        ]
        if self.response_filters:
            for col_name in self.response_filters:
                callbacks.append(Input(self.ids(f"filter-{col_name}"), "value"))
        return callbacks

    @property
    def distribution_input_callbacks(self):
        """List of Inputs for distribution callback"""
        callbacks = [
            Input(self.ids("correlation-graph"), "clickData"),
            Input(self.ids("initial-parameter"), "data"),
            Input(self.ids("ensemble"), "value"),
            Input(self.ids("responses"), "value"),
        ]
        if self.response_filters:
            for col_name in self.response_filters:
                callbacks.append(Input(self.ids(f"filter-{col_name}"), "value"))
        return callbacks

    def make_response_filters(self, filters):
        """Returns a list of active response filters"""
        filteroptions = []
        if filters:
            for i, (col_name, col_type) in enumerate(self.response_filters.items()):
                filteroptions.append(
                    {"name": col_name, "type": col_type, "values": filters[i]}
                )
        return filteroptions

    def set_callbacks(self, app):
        @app.callback(
            [
                Output(self.ids("correlation-graph"), "figure"),
                Output(self.ids("initial-parameter"), "data"),
            ],
            self.correlation_input_callbacks,
        )
        def _update_correlation_graph(ensemble, response, *filters):
            """Callback to update correlation graph

            1. Filters and aggregates response dataframe per realization
            2. Filters parameters dataframe on selected ensemble
            3. Merge parameter and response dataframe
            4. Correlate merged dataframe
            5. Sort correlation for selected response by absolute values
            6. Remove nan values return correlation graph
            """

            filteroptions = self.make_response_filters(filters)
            responsedf = filter_and_sum_responses(
                self.responsedf,
                ensemble,
                response,
                filteroptions=filteroptions,
                aggregation=self.aggregation,
            )
            parameterdf = self.parameterdf.loc[self.parameterdf["ENSEMBLE"] == ensemble]
            df = pd.merge(responsedf, parameterdf, on=["REAL"])
            corrdf = correlate(df, response=response, method=self.corr_method)
            try:
                corr_response = (
                    corrdf[response].dropna().drop(["REAL", response], axis=0)
                )

                return (
                    make_correlation_plot(
                        corr_response, response, self.plotly_theme, self.corr_method
                    ),
                    corr_response.index[-1],
                )
            except KeyError:
                return (
                    {
                        "layout": {
                            "title": "<b>Cannot calculate correlation for given selection</b><br>"
                            "Select a different response or filter setting."
                        }
                    },
                    None,
                )

        @app.callback(
            Output(self.ids("distribution-graph"), "figure"),
            self.distribution_input_callbacks,
        )
        def _update_distribution_graph(
            clickdata, initial_parameter, ensemble, response, *filters
        ):
            """Callback to update distribution graphs.

            1. Filters and aggregates response dataframe per realization
            2. Filters parameters dataframe on selected ensemble
            3. Merge parameter and response dataframe
            4. Generate scatterplot and histograms
            """
            if clickdata:
                parameter = clickdata["points"][0]["y"]
            elif initial_parameter:
                parameter = initial_parameter
            else:
                return {}
            filteroptions = self.make_response_filters(filters)
            responsedf = filter_and_sum_responses(
                self.responsedf,
                ensemble,
                response,
                filteroptions=filteroptions,
                aggregation=self.aggregation,
            )
            parameterdf = self.parameterdf.loc[self.parameterdf["ENSEMBLE"] == ensemble]
            df = pd.merge(responsedf, parameterdf, on=["REAL"])[
                ["REAL", parameter, response]
            ]
            return make_distribution_plot(df, parameter, response, self.plotly_theme)

    def add_webvizstore(self):
        if self.parameter_csv and self.response_csv:
            return [
                (
                    read_csv,
                    [
                        {
                            "csv_file": self.parameter_csv,
                        }
                    ],
                ),
                (
                    read_csv,
                    [
                        {
                            "csv_file": self.response_csv,
                        }
                    ],
                ),
            ]

        functions = [
            (
                load_parameters,
                [
                    {
                        "ensemble_paths": self.ens_paths,
                        "ensemble_set_name": "EnsembleSet",
                    }
                ],
            ),
        ]
        if self.response_file:
            functions.append(
                (
                    load_csv,
                    [
                        {
                            "ensemble_paths": self.ens_paths,
                            "csv_file": self.response_file,
                            "ensemble_set_name": "EnsembleSet",
                        }
                    ],
                )
            )
        else:
            functions.extend(self.emodel.webvizstore)
        return functions


@CACHE.memoize(timeout=CACHE.TIMEOUT)
def filter_and_sum_responses(
    dframe, ensemble, response, filteroptions=None, aggregation="sum"
):
    """Cached wrapper for _filter_and_sum_responses"""
    return _filter_and_sum_responses(
        dframe=dframe,
        ensemble=ensemble,
        response=response,
        filteroptions=filteroptions,
        aggregation=aggregation,
    )


def _filter_and_sum_responses(
    dframe,
    ensemble,
    response,
    filteroptions=None,
    aggregation="sum",
):
    """Filter response dataframe for the given ensemble
    and optional filter columns. Returns dataframe grouped and
    aggregated per realization."""

    df = dframe.copy()
    df = df.loc[df["ENSEMBLE"] == ensemble]
    if filteroptions:
        for opt in filteroptions:
            if opt["type"] == "multi" or opt["type"] == "single":
                if isinstance(opt["values"], list):
                    df = df.loc[df[opt["name"]].isin(opt["values"])]
                else:
                    if opt["name"] == "DATE" and isinstance(opt["values"], str):
                        df = df.loc[df["DATE"].astype(str) == opt["values"]]
                    else:
                        df = df.loc[df[opt["name"]] == opt["values"]]

            elif opt["type"] == "range":
                df = df.loc[
                    (df[opt["name"]] >= np.min(opt["values"]))
                    & (df[opt["name"]] <= np.max(opt["values"]))
                ]
    if aggregation == "sum":
        return df.groupby("REAL").sum().reset_index()[["REAL", response]]
    if aggregation == "mean":
        return df.groupby("REAL").mean().reset_index()[["REAL", response]]
    raise ValueError(
        f"Aggregation of response file specified as '{aggregation}'' is invalid. "
    )


@CACHE.memoize(timeout=CACHE.TIMEOUT)
def correlate(inputdf, response, method="pearson"):
    """Cached wrapper for _correlate"""
    return _correlate(inputdf=inputdf, response=response, method=method)


def _correlate(inputdf, response, method="pearson"):
    """Returns the correlation matrix for a dataframe"""
    if method == "pearson":
        corrdf = inputdf.corr(method=method)
    elif method == "spearman":
        corrdf = inputdf.rank().corr(method="pearson")
    else:
        raise ValueError(
            f"Correlation method {method} is invalid. "
            "Available methods are 'pearson' and 'spearman'"
        )
    return corrdf.reindex(corrdf[response].abs().sort_values().index)


def make_correlation_plot(series, response, theme, corr_method):
    """Make Plotly trace for correlation plot"""
    layout = theme_layout(
        theme,
        {
            "barmode": "relative",
            "margin": {"l": 200, "r": 50, "b": 20, "t": 100},
            "height": 750,
            "xaxis": {"range": [-1, 1]},
            "title": f"Correlations ({corr_method}) between {response} and input parameters",
        },
    )
    layout["font"].update({"size": 8})

    return {
        "data": [
            {"x": series.values, "y": series.index, "orientation": "h", "type": "bar"}
        ],
        "layout": layout,
    }


def make_distribution_plot(df, parameter, response, theme):
    """Make plotly traces for scatterplot and histograms for selected
    response and input parameter"""

    real_text = [f"Realization:{r}" for r in df["REAL"]]
    fig = make_subplots(
        rows=4,
        cols=2,
        specs=[
            [{"colspan": 2, "rowspan": 2}, None],
            [None, None],
            [{"rowspan": 2}, {"rowspan": 2}],
            [None, None],
        ],
    )
    fig.add_trace(
        {
            "type": "scatter",
            "showlegend": False,
            "mode": "markers",
            "x": df[parameter],
            "y": df[response],
            "text": real_text,
        },
        1,
        1,
    )
    fig.add_trace(
        {
            "type": "histogram",
            "x": df[parameter],
            "showlegend": False,
        },
        3,
        1,
    )
    fig.add_trace(
        {
            "type": "histogram",
            "x": df[response],
            "showlegend": False,
        },
        3,
        2,
    )
    fig["layout"].update(
        theme_layout(
            theme,
            {
                "height": 800,
                "bargap": 0.05,
                "xaxis": {
                    "title": parameter,
                },
                "yaxis": {"title": response},
                "xaxis2": {"title": parameter},
                "xaxis3": {"title": response},
                "title": f"Distribution of {response} and {parameter}",
            },
        )
    )
    fig["layout"]["font"].update({"size": 8})
    return fig


def make_range_slider(domid, values, col_name):
    try:
        values.apply(pd.to_numeric, errors="raise")
    except ValueError as exc:
        raise ValueError(
            f"Cannot calculate filter range for {col_name}. "
            "Ensure that it is a numerical column."
        ) from exc
    return dcc.RangeSlider(
        id=domid,
        min=values.min(),
        max=values.max(),
        step=calculate_slider_step(
            min_value=values.min(),
            max_value=values.max(),
            steps=len(list(values.unique())) - 1,
        ),
        value=[values.min(), values.max()],
        marks={
            str(values.min()): {"label": f"{values.min():.2f}"},
            str(values.max()): {"label": f"{values.max():.2f}"},
        },
    )


def theme_layout(theme, specific_layout):
    layout = {}
    layout.update(theme["layout"])
    layout.update(specific_layout)
    return layout


@CACHE.memoize(timeout=CACHE.TIMEOUT)
@webvizstore
def read_csv(csv_file) -> pd.DataFrame:
    return pd.read_csv(csv_file, index_col=False)



from typing import Dict, List, Union

import pandas as pd
from webviz_config.common_cache import CACHE


def interpolate_depth(df: pd.DataFrame) -> pd.DataFrame:
    df = (
        df.pivot_table(index=["DEPTH"], columns=["REAL"], values="PRESSURE")
        .interpolate(limit_direction="both")
        .stack("REAL")
    )
    return df.to_frame().rename(columns={0: "PRESSURE"}).reset_index()


@CACHE.memoize(timeout=CACHE.TIMEOUT)
def filter_frame(
    dframe: pd.DataFrame, column_values: Dict[str, Union[List[str], str]]
) -> pd.DataFrame:
    df = dframe.copy()
    for column, value in column_values.items():
        if isinstance(value, list):
            df = df.loc[df[column].isin(value)]
        else:
            df = df.loc[df[column] == value]
    return df



from typing import Optional, List, Dict

import pandas as pd
import numpy as np
from webviz_config.common_cache import CACHE


@CACHE.memoize(timeout=CACHE.TIMEOUT)
def filter_and_sum_responses(
    dframe: pd.DataFrame,
    ensemble: str,
    response: str,
    filteroptions: Optional[List[Dict[str, str]]] = None,
    aggregation: str = "sum",
) -> pd.DataFrame:
    """Cached wrapper for _filter_and_sum_responses"""
    return _filter_and_sum_responses(
        dframe=dframe,
        ensemble=ensemble,
        response=response,
        filteroptions=filteroptions,
        aggregation=aggregation,
    )


def _filter_and_sum_responses(
    dframe: pd.DataFrame,
    ensemble: str,
    response: str,
    filteroptions: Optional[List[Dict[str, str]]] = None,
    aggregation: str = "sum",
) -> pd.DataFrame:
    """Filter response dataframe for the given ensemble
    and optional filter columns. Returns dataframe grouped and
    aggregated per realization.
    """
    df = dframe.copy()
    df = df.loc[df["ENSEMBLE"] == ensemble]
    if filteroptions:
        for opt in filteroptions:
            if opt["type"] == "multi" or opt["type"] == "single":
                if isinstance(opt["values"], list):
                    df = df.loc[df[opt["name"]].isin(opt["values"])]
                else:
                    if opt["name"] == "DATE" and isinstance(opt["values"], str):
                        df = df.loc[df["DATE"].astype(str) == opt["values"]]
                    else:
                        df = df.loc[df[opt["name"]] == opt["values"]]

            elif opt["type"] == "range":
                df = df.loc[
                    (df[opt["name"]] >= np.min(opt["values"]))
                    & (df[opt["name"]] <= np.max(opt["values"]))
                ]
    if aggregation == "sum":
        return df.groupby("REAL").sum().reset_index()[["REAL", response]]
    if aggregation == "mean":
        return df.groupby("REAL").mean().reset_index()[["REAL", response]]
    raise ValueError(f"Unknown aggregation '{aggregation}'.")



"""These are plugins relevant within subsurface workflows. Most of them
rely on the setting `scratch_ensembles` within
`shared_settings`. This setting connects ensemble names (user defined)
with the paths to where the ensembles are stored, either absolute or
relative to the location of the configuration file.
I.e. you could have
```yaml
title: Reek Webviz Demonstration
shared_settings:
  scratch_ensembles:
    iter-0: /scratch/my_ensemble/realization-*/iter-0
    iter-1: /scratch/my_ensemble/realization-*/iter-1
pages:
  - title: Front page
    content:
      - plugin: ReservoirSimulationTimeSeries
        ensembles:
          - iter-0
          - iter-1
```
"""

from ._bhp_qc import BhpQc
from ._disk_usage import DiskUsage
from ._history_match import HistoryMatch
from ._horizon_uncertainty_viewer import HorizonUncertaintyViewer
from ._inplace_volumes import InplaceVolumes
from ._inplace_volumes_onebyone import InplaceVolumesOneByOne
from ._morris_plot import MorrisPlot
from ._parameter_analysis import ParameterAnalysis
from ._parameter_correlation import ParameterCorrelation
from ._parameter_distribution import ParameterDistribution
from ._parameter_parallel_coordinates import ParameterParallelCoordinates
from ._parameter_response_correlation import ParameterResponseCorrelation
from ._property_statistics import PropertyStatistics
from ._pvt_plot import PvtPlot
from ._relative_permeability import RelativePermeability
from ._reservoir_simulation_timeseries import ReservoirSimulationTimeSeries
from ._reservoir_simulation_timeseries_onebyone import (
    ReservoirSimulationTimeSeriesOneByOne,
)
from ._reservoir_simulation_timeseries_regional import (
    ReservoirSimulationTimeSeriesRegional,
)
from ._rft_plotter.rft_plotter import RftPlotter
from ._running_time_analysis_fmu import RunningTimeAnalysisFMU
from ._segy_viewer import SegyViewer
from ._subsurface_map import SubsurfaceMap
from ._surface_viewer_fmu import SurfaceViewerFMU
from ._surface_with_grid_cross_section import SurfaceWithGridCrossSection
from ._surface_with_seismic_cross_section import SurfaceWithSeismicCrossSection
from ._well_cross_section import WellCrossSection
from ._well_cross_section_fmu import WellCrossSectionFMU
from ._assisted_history_matching_analysis import AssistedHistoryMatchingAnalysis



from typing import List, Dict, Any, Union, Tuple, Callable
import json
from uuid import uuid4
from pathlib import Path

import dash
from dash.exceptions import PreventUpdate
from dash.dependencies import Input, Output, State
import dash_html_components as html
import dash_core_components as dcc
import webviz_core_components as wcc
from webviz_config import WebvizPluginABC
from webviz_config.webviz_store import webvizstore
from webviz_config.utils import calculate_slider_step
import numpy as np

from .._datainput.seismic import load_cube_data, get_iline, get_xline, get_zslice


class SegyViewer(WebvizPluginABC):
    """Inspired by [SegyViewer for Python](https://github.com/equinor/segyviewer) this plugin
visualizes seismic 3D cubes with 3 plots (inline, crossline and zslice).
The plots are linked and updates are done by clicking in the plots.

---

* **`segyfiles`:** List of file paths to `SEGY` files (absolute or relative to config file).
* **`zunit`:** z-unit for display.
* **`colors`:** List of hex colors use. \
Note that apostrophies should be used to avoid that hex colors are read as comments. E.g. \
`'#000000'` for black.

---

* [Examples of segyfiles](https://github.com/equinor/webviz-subsurface-testdata/tree/master/\
observed_data/seismic).

The segyfiles are on a `SEG-Y` format and can be investigated outside `webviz` using \
e.g. [xtgeo](https://xtgeo.readthedocs.io/en/latest/).

"""

    def __init__(
        self,
        app: dash.Dash,
        segyfiles: List[Path],
        zunit: str = "depth (m)",
        colors: list = None,
    ):

        super().__init__()

        self.zunit = zunit
        self.segyfiles: List[str] = [str(segy) for segy in segyfiles]
        self.initial_colors = (
            colors
            if colors
            else [
                "#67001f",
                "#ab152a",
                "#d05546",
                "#ec9372",
                "#fac8ac",
                "#faeae1",
                "#e6eff4",
                "#bbd9e9",
                "#7db7d6",
                "#3a8bbf",
                "#1f61a5",
                "#053061",
            ]
        )
        self.init_state = self.update_state(self.segyfiles[0])
        self.init_state.get("colorscale", self.initial_colors)
        self.init_state.get("uirevision", str(uuid4()))
        self.uid = uuid4()
        self.plotly_theme = app.webviz_settings["theme"].plotly_theme
        self.set_callbacks(app)

    def ids(self, element: str) -> str:
        """Generate unique id for dom element"""
        return f"{element}-id-{self.uid}"

    def update_state(self, cubepath: str, **kwargs: Any) -> Dict[str, Any]:
        cube = load_cube_data(get_path(cubepath))
        state = {
            "cubepath": cubepath,
            "iline": int(cube.ilines[int(len(cube.ilines) / 2)]),
            "xline": int(cube.xlines[int(len(cube.xlines) / 2)]),
            "zslice": float(cube.zslices[int(len(cube.zslices) / 2)]),
            "min_value": float(f"{round(cube.values.min(), 2):2f}"),
            "max_value": float(f"{round(cube.values.max(), 2):2f}"),
            "color_min_value": float(f"{round(cube.values.min(), 2):2f}"),
            "color_max_value": float(f"{round(cube.values.max(), 2):2f}"),
            "uirevision": str(uuid4()),
            "colorscale": self.initial_colors,
        }
        if kwargs:
            for key, value in kwargs.items():
                state[key] = value
        return state

    @property
    def tour_steps(self) -> List[dict]:
        return [
            {
                "id": self.ids("layout"),
                "content": (
                    "Visualizes SEG-Y cubes. Display different slices "
                    "of the cube by clicking (MB1) in the different plots. "
                    "To zoom, hold MB1 and draw a vertical/horizontal "
                    "line or a rectangle."
                ),
            },
            {
                "id": self.ids("cube"),
                "content": "The currently visualized seismic cube.",
            },
            {
                "id": self.ids("inline"),
                "content": "Selected inline for the seismic cube.",
            },
            {
                "id": self.ids("xline"),
                "content": "Selected crossline for the seismic cube.",
            },
            {
                "id": self.ids("zslice"),
                "content": "Selected zslice for the seismic cube.",
            },
            {
                "id": self.ids("color-scale"),
                "content": "Click this button to change colorscale",
            },
            {
                "id": self.ids("color-values"),
                "content": "Drag either node of slider to truncate color ranges.",
            },
            {
                "id": self.ids("color-reset"),
                "content": (
                    "Click this button to update color slider min/max and reset ranges."
                ),
            },
            {
                "id": self.ids("zoom"),
                "content": "Click this button to reset zoom/pan state in the plot.",
            },
        ]

    @property
    def settings_layout(self) -> wcc.FlexBox:
        """Layout for color and other settings"""
        return wcc.FlexBox(
            style={"margin": "50px"},
            children=[
                html.Div(
                    style={"width": "100%"},
                    children=html.Label(
                        children=[
                            html.Span("Seismic cube:", style={"font-weight": "bold"}),
                            dcc.Dropdown(
                                id=self.ids("cube"),
                                options=[
                                    {"label": Path(cube).stem, "value": cube}
                                    for cube in self.segyfiles
                                ],
                                value=self.segyfiles[0],
                                clearable=False,
                                persistence=True,
                                persistence_type="session",
                            ),
                        ]
                    ),
                ),
                html.Div(
                    style={"marginRight": "50px", "width": "50%", "marginLeft": "50px"},
                    children=[
                        html.Label(
                            style={"font-weight": "bold", "textAlign": "center"},
                            children="Set colorscale",
                        ),
                        wcc.ColorScales(
                            id=self.ids("color-scale"),
                            colorscale=self.initial_colors,
                            nSwatches=12,
                        ),
                    ],
                ),
                html.Div(
                    style={"marginRight": "50px", "width": "50%", "marginLeft": "50px"},
                    children=[
                        html.P(
                            "Color range",
                            style={"textAlign": "center", "font-weight": "bold"},
                        ),
                        dcc.RangeSlider(
                            id=self.ids("color-values"),
                            min=self.init_state["min_value"],
                            max=self.init_state["max_value"],
                            value=[
                                self.init_state["min_value"],
                                self.init_state["max_value"],
                            ],
                            tooltip={"always_visible": True},
                            step=calculate_slider_step(
                                min_value=self.init_state["min_value"],
                                max_value=self.init_state["max_value"],
                                steps=100,
                            ),
                            persistence=True,
                            persistence_type="session",
                        ),
                    ],
                ),
                html.Button(id=self.ids("color-reset"), children="Reset Range"),
                html.Button(id=self.ids("zoom"), children="Reset zoom"),
            ],
        )

    @property
    def layout(self) -> html.Div:
        return html.Div(
            id=self.ids("layout"),
            children=wcc.FlexBox(
                children=[
                    html.Div(
                        style={"width": "50%"},
                        children=[
                            html.Div(
                                style={"minWidth": "200px", "height": "400px"},
                                children=wcc.Graph(
                                    config={"displayModeBar": False},
                                    id=self.ids("inline"),
                                ),
                            ),
                            html.Div(
                                style={"minWidth": "200px", "height": "400px"},
                                children=wcc.Graph(
                                    config={"displayModeBar": False},
                                    id=self.ids("xline"),
                                ),
                            ),
                        ],
                    ),
                    html.Div(
                        style={"width": "50%"},
                        children=[
                            html.Div(
                                style={"minWidth": "200px", "height": "400px"},
                                children=wcc.Graph(
                                    config={"displayModeBar": False},
                                    id=self.ids("zslice"),
                                ),
                            ),
                            self.settings_layout,
                        ],
                    ),
                    dcc.Store(
                        id=self.ids("state-storage"),
                        storage_type="session",
                        data=json.dumps(self.init_state),
                    ),
                ]
            ),
        )

    # pylint: disable=too-many-statements
    def set_callbacks(self, app: dash.Dash) -> None:
        @app.callback(
            Output(self.ids("state-storage"), "data"),
            [
                Input(self.ids("cube"), "value"),
                Input(self.ids("inline"), "clickData"),
                Input(self.ids("xline"), "clickData"),
                Input(self.ids("zslice"), "clickData"),
                Input(self.ids("color-values"), "value"),
                Input(self.ids("color-scale"), "colorscale"),
                Input(self.ids("zoom"), "n_clicks"),
                Input(self.ids("color-reset"), "n_clicks"),
            ],
            [
                State(self.ids("zslice"), "figure"),
                State(self.ids("state-storage"), "data"),
            ],
        )
        # pylint: disable=unused-argument,too-many-arguments
        def _update_state(
            cubepath: str,
            icd: Union[dict, None],
            xcd: Union[dict, None],
            zcd: Union[dict, None],
            color_values: List[float],
            colorscale: List[float],
            _zoom_btn: Union[int, None],
            _reset_range_btn: Union[int, None],
            zfig: Union[dict, None],
            state_data_str: str,
        ) -> str:
            """Updates dcc.Store object with active iline, xline, zlayer and color settings."""

            store: dict = json.loads(state_data_str)
            ctx = dash.callback_context.triggered[0]["prop_id"]
            x_was_clicked = (
                xcd if xcd and ctx == f"{self.ids('xline')}.clickData" else None
            )
            i_was_clicked = (
                icd if icd and ctx == f"{self.ids('inline')}.clickData" else None
            )
            z_was_clicked = (
                zcd if zcd and ctx == f"{self.ids('zslice')}.clickData" else None
            )
            if ctx == f"{self.ids('cube')}.value":
                store = self.update_state(cubepath)
            if ctx == f"{self.ids('zoom')}.n_clicks":
                store["uirevision"] = str(uuid4())
            if x_was_clicked and xcd is not None:
                store["iline"] = xcd["points"][0]["x"]
                store["zslice"] = xcd["points"][0]["y"]
            if z_was_clicked and zcd is not None and zfig is not None:
                store["iline"] = zcd["points"][0]["x"]
                store["xline"] = zcd["points"][0]["y"]
                store["zslice"] = float(zfig["data"][0]["text"])
            if i_was_clicked and icd is not None:
                store["xline"] = icd["points"][0]["x"]
                store["zslice"] = icd["points"][0]["y"]
            if ctx == f"{self.ids('color-reset')}.n_clicks":
                store["color_min_value"] = store["min_value"]
                store["color_max_value"] = store["max_value"]
            else:
                store["color_min_value"] = color_values[0]
                store["color_max_value"] = color_values[1]
            store["colorscale"] = colorscale
            return json.dumps(store)

        @app.callback(
            Output(self.ids("zslice"), "figure"),
            [Input(self.ids("state-storage"), "data")],
        )
        def _set_zslice(state_data_str: Union[str, None]) -> dict:
            """Updates z-slice heatmap"""
            if not state_data_str:
                raise PreventUpdate
            state = json.loads(state_data_str)
            cube = load_cube_data(get_path(state["cubepath"]))
            shapes = [
                {
                    "type": "line",
                    "x0": state["iline"],
                    "y0": cube.xlines[0],
                    "x1": state["iline"],
                    "y1": cube.xlines[-1],
                    "line": {"width": 1, "dash": "dot"},
                },
                {
                    "type": "line",
                    "y0": state["xline"],
                    "x0": cube.ilines[0],
                    "y1": state["xline"],
                    "x1": cube.ilines[-1],
                    "line": {"width": 1, "dash": "dot"},
                },
            ]

            zslice_arr = get_zslice(cube, state["zslice"])

            fig = make_heatmap(
                zslice_arr,
                self.plotly_theme,
                showscale=True,
                text=str(state["zslice"]),
                title=f'Zslice {state["zslice"]} ({self.zunit})',
                xaxis_title="Inline",
                yaxis_title="Crossline",
                reverse_y=False,
                zmin=state["color_min_value"],
                zmax=state["color_max_value"],
                colorscale=state["colorscale"],
                uirevision=state["uirevision"],
            )
            fig["layout"]["shapes"] = shapes
            fig["layout"]["xaxis"].update({"constrain": "domain"})
            fig["layout"]["yaxis"].update({"scaleanchor": "x"})
            return fig

        @app.callback(
            Output(self.ids("inline"), "figure"),
            [Input(self.ids("state-storage"), "data")],
        )
        def _set_iline(state_data_str: Union[str, None]) -> dict:
            """Updates inline heatmap"""
            if not state_data_str:
                raise PreventUpdate
            state = json.loads(state_data_str)
            cube = load_cube_data(get_path(state["cubepath"]))
            shapes = [
                {
                    "type": "line",
                    "x0": state["xline"],
                    "y0": cube.zslices[0],
                    "x1": state["xline"],
                    "y1": cube.zslices[-1],
                    "line": {"width": 1, "dash": "dot"},
                },
                {
                    "type": "line",
                    "x0": cube.xlines[0],
                    "y0": state["zslice"],
                    "x1": cube.xlines[-1],
                    "y1": state["zslice"],
                    "line": {"width": 1, "dash": "dot"},
                },
            ]
            iline_arr = get_iline(cube, state["iline"])

            fig = make_heatmap(
                iline_arr,
                self.plotly_theme,
                xaxis=cube.xlines,
                yaxis=cube.zslices,
                title=f'Inline {state["iline"]}',
                xaxis_title="Crossline",
                yaxis_title=self.zunit,
                zmin=state["color_min_value"],
                zmax=state["color_max_value"],
                colorscale=state["colorscale"],
                uirevision=state["uirevision"],
            )
            fig["layout"]["shapes"] = shapes
            return fig

        @app.callback(
            Output(self.ids("xline"), "figure"),
            [Input(self.ids("state-storage"), "data")],
        )
        def _set_xline(state_data_str: Union[str, None]) -> dict:
            """Update xline heatmap"""
            if not state_data_str:
                raise PreventUpdate
            state = json.loads(state_data_str)
            cube = load_cube_data(get_path(state["cubepath"]))
            shapes = [
                {
                    "type": "line",
                    "x0": state["iline"],
                    "y0": cube.zslices[0],
                    "x1": state["iline"],
                    "y1": cube.zslices[-1],
                    "line": {"width": 1, "dash": "dot"},
                },
                {
                    "type": "line",
                    "x0": cube.ilines[0],
                    "y0": state["zslice"],
                    "x1": cube.ilines[-1],
                    "y1": state["zslice"],
                    "line": {"width": 1, "dash": "dot"},
                },
            ]
            xline_arr = get_xline(cube, state["xline"])
            fig = make_heatmap(
                xline_arr,
                self.plotly_theme,
                xaxis=cube.ilines,
                yaxis=cube.zslices,
                title=f'Crossline {state["xline"]}',
                xaxis_title="Inline",
                yaxis_title=self.zunit,
                zmin=state["color_min_value"],
                zmax=state["color_max_value"],
                colorscale=state["colorscale"],
                uirevision=state["uirevision"],
            )
            fig["layout"]["shapes"] = shapes
            return fig

        @app.callback(
            [
                Output(self.ids("color-values"), "min"),
                Output(self.ids("color-values"), "max"),
                Output(self.ids("color-values"), "value"),
                Output(self.ids("color-values"), "step"),
            ],
            [Input(self.ids("color-reset"), "n_clicks")],
            [State(self.ids("state-storage"), "data")],
        )
        # pylint: disable=unused-argument
        def _reset_color_range(
            _btn_clicked: int, state_data_str: Union[str, None]
        ) -> Tuple[float, float, list, float]:
            if not state_data_str:
                raise PreventUpdate
            state = json.loads(state_data_str)
            minv = state["min_value"]
            maxv = state["max_value"]
            value = [minv, maxv]
            step = calculate_slider_step(min_value=minv, max_value=maxv, steps=100)
            return minv, maxv, value, step

    @staticmethod
    def set_grid_layout(columns: str) -> dict:
        return {
            "display": "grid",
            "alignContent": "space-around",
            "justifyContent": "space-between",
            "gridTemplateColumns": f"{columns}",
        }

    def add_webvizstore(self) -> List[Tuple[Callable, list]]:
        return [(get_path, [{"path": fn}]) for fn in self.segyfiles]


# pylint: disable=too-many-arguments
# pylint: disable=too-many-locals
def make_heatmap(
    arr: np.ndarray,
    theme: dict,
    height: int = 400,
    zmin: float = None,
    zmax: float = None,
    colorscale: list = None,
    uirevision: str = None,
    showscale: bool = False,
    reverse_y: bool = True,
    xaxis: np.array = None,
    yaxis: np.array = None,
    text: str = None,
    title: str = None,
    yaxis_title: str = None,
    xaxis_title: str = None,
) -> Dict[str, Any]:
    """Createst heatmap plot"""
    colors = (
        [[i / (len(colorscale) - 1), color] for i, color in enumerate(colorscale)]
        if colorscale
        else "RdBu"
    )
    layout = {}
    layout.update(theme["layout"])
    layout.update(
        {
            "height": height,
            "title": title,
            "uirevision": uirevision,
            "margin": {"b": 50, "t": 50, "r": 0},
            "yaxis": {
                "title": yaxis_title,
                "autorange": "reversed" if reverse_y else None,
            },
            "xaxis": {"title": xaxis_title},
        }
    )

    return {
        "data": [
            {
                "type": "heatmap",
                "text": text if text else None,
                "z": arr.tolist(),
                "x": xaxis,
                "y": yaxis,
                "zsmooth": "best",
                "showscale": showscale,
                "colorscale": colors,
                "zmin": zmin,
                "zmax": zmax,
            }
        ],
        "layout": layout,
    }


@webvizstore
def get_path(path: str) -> Path:
    return Path(path)



from typing import Tuple, Union, TYPE_CHECKING

import numpy as np
import pandas as pd
from dash.dependencies import Input, Output, State, ALL
from dash.exceptions import PreventUpdate
import dash

from ..utils.colors import find_intermediate_color_rgba
from ..figures.correlation_figure import CorrelationFigure
from ..utils.surface import surface_from_zone_prop
from ..models import SimulationTimeSeriesModel
from ..models.surface_leaflet_model import (
    SurfaceLeafletModel,
)

if TYPE_CHECKING:
    # pylint: disable=cyclic-import
    from ..property_statistics import PropertyStatistics


def property_response_controller(parent: "PropertyStatistics", app: dash.Dash) -> None:
    @app.callback(
        Output({"id": parent.uuid("surface-view"), "tab": "response"}, "layers"),
        Output({"id": parent.uuid("surface-name"), "tab": "response"}, "children"),
        Input(parent.uuid("property-response-correlation-graph"), "clickData"),
        Input({"id": parent.uuid("ensemble-selector"), "tab": "response"}, "value"),
        Input({"id": parent.uuid("surface-type"), "tab": "response"}, "value"),
    )
    def _update_surface(
        clickdata: Union[None, dict], ensemble: str, stype: str
    ) -> Tuple[list, str]:
        if clickdata is not None:
            label = clickdata["points"][0]["y"]
            prop = label.split(" | ")[0]
            zone = label.split(" | ")[1]
            sprop = parent.surface_renaming.get(prop, prop)
            szone = parent.surface_renaming.get(zone, zone)
            surface = surface_from_zone_prop(
                parent, zone=szone, prop=sprop, ensemble=ensemble, stype=stype
            )
            surface_layer = SurfaceLeafletModel(surface, name="surface").layer
            return [surface_layer], f"{stype.capitalize()} for {prop}, {zone}"
        raise PreventUpdate

    @app.callback(
        Output(parent.uuid("property-response-correlated-slider"), "min"),
        Output(parent.uuid("property-response-correlated-slider"), "max"),
        Output(parent.uuid("property-response-correlated-slider"), "step"),
        Output(parent.uuid("property-response-correlated-slider"), "value"),
        Output(parent.uuid("property-response-correlated-slider"), "marks"),
        Output(parent.uuid("property-response-correlated-slider"), "disabled"),
        Input(parent.uuid("property-response-correlated-filter"), "value"),
        Input({"id": parent.uuid("ensemble-selector"), "tab": "response"}, "value"),
    )
    # pylint: disable=protected-access
    def _update_correlation_figure(
        label: str, ensemble: str
    ) -> Tuple[
        Union[float, dash.dash._NoUpdate],
        Union[float, dash.dash._NoUpdate],
        Union[float, dash.dash._NoUpdate],
        list,
        Union[dict, dash.dash._NoUpdate],
        bool,
    ]:
        if label is not None:
            values = parent.pmodel.filter_on_label(label=label, ensemble=ensemble)
            return (
                values.min(),
                values.max(),
                (values.max() - values.min()) / 100,
                [values.min(), values.max()],
                {
                    str(values.min()): {"label": f"{values.min():.2f}"},
                    str(values.max()): {"label": f"{values.max():.2f}"},
                },
                False,
            )

        return (
            dash.no_update,
            dash.no_update,
            dash.no_update,
            [None, None],
            dash.no_update,
            True,
        )

    @app.callback(
        Output(parent.uuid("property-response-vector-graph"), "figure"),
        Output(parent.uuid("property-response-correlation-graph"), "figure"),
        Input({"id": parent.uuid("ensemble-selector"), "tab": "response"}, "value"),
        Input(parent.uuid("property-response-vector-select"), "value"),
        Input(parent.uuid("property-response-vector-graph"), "clickData"),
        Input(parent.uuid("property-response-correlation-graph"), "clickData"),
        Input(
            {
                "id": parent.uuid("filter-selector"),
                "tab": "response",
                "selector": ALL,
            },
            "value",
        ),
        Input(parent.uuid("property-response-correlated-slider"), "value"),
        State(parent.uuid("property-response-vector-graph"), "figure"),
        State(parent.uuid("property-response-correlated-filter"), "value"),
    )
    # pylint: disable=too-many-locals
    def _update_graphs(
        ensemble: str,
        vector: str,
        timeseries_clickdata: Union[None, dict],
        correlation_clickdata: Union[None, dict],
        selectors: list,
        corr_filter: list,
        figure: dict,
        label: str,
    ) -> Tuple[dict, dict]:
        if (
            dash.callback_context.triggered is None
            or dash.callback_context.triggered[0]["prop_id"] == "."
            or vector is None
        ):
            raise PreventUpdate
        ctx = dash.callback_context.triggered[0]["prop_id"].split(".")[0]

        # Filter realizations if correlation filter is active
        real_filter = (
            parent.pmodel.filter_reals_on_label_range(ensemble, label, corr_filter)
            if corr_filter[0] is not None and corr_filter[1] is not None
            else None
        )

        # Make timeseries graph
        if any(
            substr in ctx
            for substr in [
                parent.uuid("property-response-vector-select"),
                parent.uuid("ensemble-selector"),
                parent.uuid("property-response-correlated-slider"),
            ]
        ):

            figure = update_timeseries_graph(
                parent.vmodel, ensemble, vector, real_filter
            )

        # Get clicked data or last available date initially
        date = (
            timeseries_clickdata.get("points", [{}])[0].get(
                "x", parent.vmodel.get_last_date(ensemble)
            )
            if timeseries_clickdata
            else parent.vmodel.get_last_date(ensemble)
        )

        # Draw clicked date as a black line
        ymin = min([min(trace["y"]) for trace in figure["data"]])
        ymax = max([max(trace["y"]) for trace in figure["data"]])
        figure["layout"]["shapes"] = [
            {"type": "line", "x0": date, "x1": date, "y0": ymin, "y1": ymax}
        ]

        # Get dataframe with vector and REAL
        vector_df = parent.vmodel.get_ensemble_vector_for_date(
            ensemble=ensemble, vector=vector, date=date
        )
        vector_df["REAL"] = vector_df["REAL"].astype(int)

        # Get dataframe with properties per label and REAL
        prop_df = parent.pmodel.get_ensemble_properties(ensemble, selectors)
        prop_df["REAL"] = prop_df["REAL"].astype(int)
        prop_df = (
            prop_df[prop_df["REAL"].isin(real_filter)]
            if real_filter is not None
            else prop_df
        )

        # Correlate properties against vector
        corrseries = correlate(vector_df, prop_df, response=vector)
        # Make correlation figure
        correlation_figure = CorrelationFigure(corrseries, n_rows=20, title="")

        # Get clicked correlation bar or largest bar initially
        selected_corr = (
            correlation_clickdata.get("points", [{}])[0].get("y")
            if correlation_clickdata
            else correlation_figure.first_y_value
        )

        # Update bar colors
        correlation_figure.set_bar_colors(selected_corr)

        # Order realizations sorted on value of property
        real_order = (
            parent.pmodel.get_real_order(ensemble, series=selected_corr)
            if selected_corr is not None
            else None
        )

        # Color timeseries lines from value of property
        if real_order is not None:
            mean = real_order["Avg"].mean()
            low_reals = real_order[real_order["Avg"] <= mean]["REAL"].astype(str).values
            high_reals = real_order[real_order["Avg"] > mean]["REAL"].astype(str).values
            for trace_no, trace in enumerate(figure.get("data", [])):
                if trace["name"] == ensemble:
                    figure["data"][trace_no]["marker"]["color"] = set_real_color(
                        str(trace["customdata"]), low_reals, high_reals
                    )
            figure["layout"]["title"] = f"Colored by {selected_corr}"

        return figure, correlation_figure.figure


def set_real_color(real_no: str, low_reals: list, high_reals: list) -> str:

    if real_no in low_reals:
        index = int(list(low_reals).index(real_no))
        intermed = index / len(low_reals)
        return find_intermediate_color_rgba(
            "rgba(255,0,0, 100, .1)", "rgba(220,220,220, 0.1)", intermed
        )
    if real_no in high_reals:
        index = int(list(high_reals).index(real_no))
        intermed = index / len(high_reals)
        return find_intermediate_color_rgba(
            "rgba(220,220,220, 0.1)", "rgba(50,205,50, 1)", intermed
        )

    return "rgba(220,220,220, 0.2)"


def update_timeseries_graph(
    timeseries_model: SimulationTimeSeriesModel,
    ensemble: str,
    vector: str,
    real_filter: pd.Series = None,
) -> dict:

    return {
        "data": timeseries_model.add_realization_traces(
            ensemble=ensemble, vector=vector, real_filter=real_filter
        ),
        "layout": dict(
            margin={"r": 40, "l": 40, "t": 40, "b": 40},
        ),
    }


def correlate(vectordf: pd.DataFrame, propdf: pd.DataFrame, response: str) -> pd.Series:
    """Returns the correlation matrix for a dataframe"""
    df = pd.merge(propdf, vectordf, on=["REAL"])
    df = df[df.columns[df.nunique() > 1]]
    if response not in df.columns:
        df[response] = np.nan
    series = df[response]
    df = df.drop(columns=[response, "REAL"])
    corrdf = df.corrwith(series)
    return corrdf.reindex(corrdf.abs().sort_values().index)
